from .conditional import (
    FullConditionalMean,
    FullConditionalMeanTime,
    LandmarksConditionalMean,
    LandmarksConditionalMeanTime,
    LandmarksConditionalMeanCholesky,
    LandmarksConditionalMeanCholeskyTime,
)

__all__ = [
    "FullConditionalMean",
    "FullConditionalMeanTime",
    "LandmarksConditionalMean",
    "LandmarksConditionalMeanTime",
    "LandmarksConditionalMeanCholesky",
    "LandmarksConditionalMeanCholeskyTime",
]



import jax
from jax.numpy import isscalar, atleast_2d

from .validation import _validate_1d, _validate_float


def derivative(function, x, jit=True):
    """
    Computes the derivative of a scalar function at each point in `x`.

    This function applies a jax-based derivative operation to the input function evaluated at specific points in `x`.
    The derivative is with respect to the function's input.

    Parameters
    ----------
    function : callable
        A function that takes in a scalar input and outputs a scalar.
        The function must have the signature function(x: scalar) -> scalar.
    x : array-like or scalar
        Data point or points at which to evaluate the derivative.
        If `x` is an array then the derivative will be computed for
        all points in the array. It must be 1-d.
    jit : bool, optional
        If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

    Returns
    -------
    array-like
        The derivative of the function evaluated at each point in `x`.
        The shape of the output array is the same as `x`.

    """

    def get_grad(x):
        return jax.jacrev(function)(x)

    if isscalar(x):
        x = _validate_float(x, "x")
        return get_grad(x)

    x = _validate_1d(x)

    if jit:
        get_grad = jax.jit(get_grad)
    return jax.vmap(get_grad, in_axes=(0,))(x).T


def gradient(function, x, jit=True):
    """
    Computes the gradient of a function for each line in `x`.

    Parameters
    ----------
    function : callable
        A function that takes a scalar input and outputs a scalar.
        The function must have the signature function(x: scalar) -> scalar.
    x : array-like
        Data points at which the gradient is to be computed.
    jit : bool, optional
        If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

    Returns
    -------
    array-like
        The gradient of the function at each point in `x`.
        The shape of the output array is the same as `x`.
    """

    def get_grad(x):
        return jax.jacrev(function)(x[None, :])

    if jit:
        get_grad = jax.jit(get_grad)
    return jax.vmap(get_grad, in_axes=(0,))(x).reshape(x.shape)


def hessian(function, x, jit=True):
    """
    Computes the gradient of a function for each line in `x`.

    Parameters
    ----------
    function : callable
        A function that takes a scalar input and outputs a scalar.
        The function must have the signature function(x: scalar) -> scalar.
    x : array-like
        Data points at which the gradient is to be computed.
    jit : bool, optional
        If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

    Returns
    -------
    array-like
        The gradient of the function at each point in `x`.
        The shape of the output array is the same as `x`.
    """
    x = atleast_2d(x)

    def get_hess(x):
        return jax.jacfwd(jax.jacrev(function))(x[None, :])

    if jit:
        get_hess = jax.jit(get_hess)
    out_shape = x.shape + x.shape[1:]
    return jax.vmap(get_hess, in_axes=(0,))(x).reshape(out_shape)


def hessian_log_determinant(function, x, jit=True):
    """
    Computes the logarithm of the determinant of the Hessian for each line in `x`.

    Parameters
    ----------
    function : callable
        A function that takes a scalar input and outputs a scalar.
        The function must have the signature function(x: scalar) -> scalar.
    x : array-like
        Data points at which the log determinant of the Hessian is to be computed.
    jit : bool, optional
        If True, use JAX's just-in-time (JIT) compilation to speed up the computation. Defaults to True.

    Returns
    -------
    array-like, array-like
        The sign of the determinant at each point in `x` and the logarithm of its absolute value.
        `signs.shape == log_determinants.shape == x.shape[0]`.
    """
    x = atleast_2d(x)

    d = x.shape[1]
    hess_shape = (d, d)

    def get_log_det(x):
        hess = jax.jacfwd(jax.jacrev(function))(x[None, :]).reshape(hess_shape)
        sign, log_det = jax.numpy.linalg.slogdet(hess)
        return sign, log_det

    if jit:
        get_log_det = jax.jit(get_log_det)
    return jax.vmap(get_log_det, in_axes=(0,))(x)



from .cov import Matern52
from .decomposition import DEFAULT_RANK, DEFAULT_METHOD
from .inference import (
    minimize_adam,
    run_advi,
    minimize_lbfgsb,
    DEFAULT_N_ITER,
    DEFAULT_INIT_LEARN_RATE,
    DEFAULT_OPTIMIZER,
    DEFAULT_JIT,
)
from .parameters import (
    compute_landmarks,
    compute_nn_distances,
    compute_ls,
    compute_cov_func,
    compute_L,
    DEFAULT_N_LANDMARKS,
)
from .util import (
    test_rank,
    DEFAULT_JITTER,
    Log,
)
from .validation import (
    _validate_positive_int,
    _validate_positive_float,
    _validate_float_or_int,
    _validate_float,
    _validate_string,
    _validate_bool,
    _validate_array,
    _validate_cov_func_curry,
    _validate_cov_func,
    _validate_float_or_iterable_numerical,
)


DEFAULT_COV_FUNC = Matern52
RANK_FRACTION_THRESHOLD = 0.8
SAMPLE_LANDMARK_RATIO = 10

logger = Log()


class BaseEstimator:
    R"""
    Base class for the mellon estimators.
    """

    def __init__(
        self,
        cov_func_curry=DEFAULT_COV_FUNC,
        n_landmarks=DEFAULT_N_LANDMARKS,
        rank=DEFAULT_RANK,
        method=DEFAULT_METHOD,
        jitter=DEFAULT_JITTER,
        optimizer=DEFAULT_OPTIMIZER,
        n_iter=DEFAULT_N_ITER,
        init_learn_rate=DEFAULT_INIT_LEARN_RATE,
        landmarks=None,
        nn_distances=None,
        d=None,
        mu=0,
        ls=None,
        ls_factor=1,
        cov_func=None,
        L=None,
        initial_value=None,
        jit=DEFAULT_JIT,
    ):
        self.cov_func_curry = _validate_cov_func_curry(
            cov_func_curry, cov_func, "cov_func_curry"
        )
        self.n_landmarks = _validate_positive_int(n_landmarks, "n_landmarks")
        self.rank = _validate_float_or_int(rank, "rank")
        self.method = _validate_string(
            method, "method", choices={"percent", "fixed", "auto"}
        )
        self.jitter = _validate_positive_float(jitter, "jitter")
        self.landmarks = _validate_array(landmarks, "landmarks", optional=True)
        self.nn_distances = _validate_array(nn_distances, "nn_distances", optional=True)
        self.mu = _validate_float(mu, "mu", optional=True)
        self.ls = _validate_positive_float(ls, "ls", optional=True)
        self.ls_factor = _validate_positive_float(ls_factor, "ls_factor")
        self.cov_func = _validate_cov_func(cov_func, "cov_func", optional=True)
        self.L = _validate_array(L, "L", optional=True)
        self.d = _validate_float_or_iterable_numerical(d, "d", optional=True)
        self.initial_value = _validate_array(
            initial_value, "initial_value", optional=True
        )
        self.optimizer = _validate_string(
            optimizer, "optimizer", choices={"adam", "advi", "L-BFGS-B"}
        )
        self.n_iter = _validate_positive_int(n_iter, "n_iter")
        self.init_learn_rate = _validate_positive_float(
            init_learn_rate, "init_learn_rate"
        )
        self.jit = _validate_bool(jit, "jit")
        self.x = None
        self.pre_transformation = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        name = self.__class__.__name__
        string = (
            f"{name}("
            f"cov_func_curry={self.cov_func_curry}, "
            f"n_landmarks={self.n_landmarks}, "
            f"rank={self.rank}, "
            f"jitter={self.jitter}, "
            f"landmarks={self.landmarks}, "
        )
        if self.nn_distances is None:
            string += "nn_distances=None, "
        else:
            string += "nn_distances=nn_distances, "
        string += f"mu={self.mu}, " f"ls={self.mu}, " f"cov_func={self.cov_func}, "
        if self.L is None:
            string += "L=None, "
        else:
            string += "L=L, "
        return string

    def __call__(self, x=None):
        """This calls self.fit_predict(x):
        Fit model and make prediction on training data x.

        :param x: Data points.
        :type x: array-like
        :return: Predictions.
        :rtype: array-like
        """
        return self.fit_predict(x=x)

    def set_x(self, x):
        """
        Sets the training instances (x) for the model and validates that they are
        formatted correctly.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features)
            The training instances where `n_samples` is the number of samples and `n_features`
            is the number of features. Each sample is an array of features representing a
            point in the feature space.

        Returns
        -------
        array-like of shape (n_samples, n_features)
            The validated training instances.

        Raises
        ------
        ValueError
            If the input `x` is not valid. For instance, `x` may not be valid if it is not
            a numerical array-like object, or if its shape does not match the required shape
            (n_samples, n_features).
        """
        self.x = _validate_array(x, "x")
        return self.x

    def _compute_landmarks(self):
        x = self.x
        n_landmarks = self.n_landmarks
        n_samples = x.shape[0]
        if n_samples > 100 * n_landmarks and n_samples > 1e6:
            logger.info(
                f"Large number of {n_samples:,} cells and "
                f"small number of {n_landmarks:,} landmarks. Consider "
                "computing k-means on a subset of cells and passing "
                "the results as 'landmarks' to speed up the process."
            )
        landmarks = compute_landmarks(x, n_landmarks=n_landmarks)
        return landmarks

    def _compute_nn_distances(self):
        x = self.x
        logger.info("Computing nearest neighbor distances.")
        nn_distances = compute_nn_distances(x)
        return nn_distances

    def _compute_ls(self):
        nn_distances = self.nn_distances
        ls = compute_ls(nn_distances)
        ls *= self.ls_factor
        return ls

    def _compute_cov_func(self):
        cov_func_curry = self.cov_func_curry
        ls = self.ls
        cov_func = compute_cov_func(cov_func_curry, ls)
        logger.info("Using covariance function %s.", str(cov_func))
        return cov_func

    def _compute_L(self):
        """
        This function calculates the lower triangular matrix L that is needed for
        computations involving the covariance matrix of the Gaussian Process model.
        """

        # Extract instance attributes
        x = self.x
        cov_func = self.cov_func
        landmarks = self.landmarks
        n_samples = x.shape[0]
        n_landmarks = n_samples if landmarks is None else landmarks.shape[0]
        rank = self.rank
        method = self.method
        jitter = self.jitter

        is_rank_full = (
            isinstance(rank, int)
            and rank == n_landmarks
            or isinstance(rank, float)
            and rank == 1.0
        )

        # Log the method and rank used for computation
        if not is_rank_full and method != "fixed":
            logger.info(
                f'Computing rank reduction using "{method}" method '
                f"retaining > {rank:.2%} of variance."
            )
        elif not is_rank_full:
            logger.info(
                f'Computing rank reduction to rank {rank} using "{method}" method.'
            )

        try:
            # Compute the lower triangular matrix L
            L = compute_L(
                x,
                cov_func,
                landmarks=landmarks,
                rank=rank,
                method=method,
                jitter=jitter,
            )
        except Exception as e:
            logger.error(f"Error during computation of L: {e}")
            raise

        new_rank = L.shape[1]

        # Check if the new rank is too high in comparison to the number of landmarks
        if (
            not is_rank_full
            and method != "fixed"
            and new_rank > (rank * RANK_FRACTION_THRESHOLD * n_landmarks)
        ):
            logger.warning(
                f"Shallow rank reduction from {n_landmarks:,} to {new_rank:,} "
                "indicates underrepresentation by landmarks. Consider "
                "increasing n_landmarks!"
            )

        # Check if the number of landmarks is sufficient for the number of samples
        if (
            is_rank_full
            and n_landmarks is not None
            and SAMPLE_LANDMARK_RATIO * n_landmarks < n_samples
        ):
            logger.info(
                "Estimating approximation accuracy "
                f"since {n_samples:,} samples are more than {SAMPLE_LANDMARK_RATIO} x "
                f"{n_landmarks:,} landmarks."
            )
            test_rank(L, threshold=RANK_FRACTION_THRESHOLD)
        logger.info(f"Using rank {new_rank:,} covariance representation.")
        return L

    def _run_inference(self):
        function = self.loss_func
        initial_value = self.initial_value
        n_iter = self.n_iter
        init_learn_rate = self.init_learn_rate
        optimizer = self.optimizer
        logger.info("Running inference using %s.", optimizer)
        if optimizer == "adam":
            results = minimize_adam(
                function,
                initial_value,
                n_iter=n_iter,
                init_learn_rate=init_learn_rate,
                jit=self.jit,
            )
            self.pre_transformation = results.pre_transformation
            self.opt_state = results.opt_state
            self.losses = results.losses
        elif optimizer == "advi":
            results = run_advi(
                function,
                initial_value,
                n_iter=n_iter,
                init_learn_rate=init_learn_rate,
                jit=self.jit,
            )
            self.pre_transformation = results.pre_transformation
            self.pre_transformation_std = results.pre_transformation_std
            self.losses = results.losses
        elif optimizer == "L-BFGS-B":
            results = minimize_lbfgsb(
                function,
                initial_value,
                jit=self.jit,
            )
            self.pre_transformation = results.pre_transformation
            self.opt_state = results.opt_state
            self.losses = [
                results.loss,
            ]
        else:
            error = ValueError(
                f"Unknown optimizer {optimizer}. You can use .loss_func and "
                ".initial_value as loss function and initial state for an "
                "external optimization. Write optimal state to "
                ".pre_transformation to enable prediction with .predict()."
            )
            logger.error(error)
            raise error

    def _prepare_attribute(self, attribute):
        R"""
        If self.attribute is None, sets self.attribute to the value of its
        corresponding _compute_attribute function. If self.attribute is None, does nothing.

        :param attribute: The name of the attribute.
        :type attribute: str
        """
        if getattr(self, attribute) is not None:
            return
        function_name = "_compute_" + attribute
        function = getattr(self, function_name)
        value = function()
        setattr(self, attribute, value)

    def prepare_inference(self, x):
        R"""
        Set all attributes in preparation for fitting. It is not necessary
        to call this function before calling fit.
        """
        ...

    def fit(self):
        R"""
        Fit the model.
        """
        ...

    @property
    def predict(self):
        R"""
        Make prediction for new data x.

        :param x: Data points.
        :type x: array-like
        :return: Predictions.
        :rtype: array-like
        """
        ...

    def fit_predict(self, x):
        R"""
        Fit model and make prediction on training data x.

        :param x: Data points.
        :type x: array-like
        :return: Predictions.
        :rtype: array-like
        """
        ...



from jax.numpy import exp, unique, corrcoef, zeros, abs, stack
from jax.numpy import sum as arraysum
from jax.numpy.linalg import norm
from jaxopt import ScipyMinimize
from .density_estimator import DensityEstimator
from .util import Log
from .validation import _validate_time_x

logger = Log()


def compute_ls_time(
    nn_distances,
    x,
    cov_func_curry,
    times=None,
    warn_below=500,
    return_data=False,
    density_estimator_kwargs=dict(),
):
    """
    Compute the optimal length scale for time (`ls_time`) based on density
    estimates and correlations at each unique time point.

    Parameters
    ----------
    nn_distances : array-like
        The nearest neighbor distances.
    x : array-like
        The training instances where the last column encodes the time point for each instance.
        If 'times' is None, the last column of 'x' is interpreted as the times.
        Shape must be (n_samples, n_features).
    times : array-like, optional
        An array encoding the time points associated with each cell/row in 'x'.
        If provided, it overrides the last column of 'x' as the times.
        Shape must be either (n_samples,) or (n_samples, 1).
    cov_func_curry : function
        The covariance function curry.
    warn_below : int, optional
        The lower limit for the number of cells at a specific time point, below which a warning is issued.
    return_data : bool, optional
        Whether to return additional data along with `ls_time`. Defaults to False.
    density_estimator_kwargs : dict, optional
        Keyword arguments passed to the `DensityEstimator`. Defaults to an empty dict.

    Returns
    -------
    ls : float
        The optimal length scale for time (`ls_time`).
    densities : array-like, optional
        The estimated densities for each training instance at each unique time point. Only returned if `return_data` is True.
    predictors : list of `DensityEstimator` instances, optional
        The `DensityEstimator` instances used to estimate the densities. Only returned if `return_data` is True.
    unique_times : array-like, optional
        The unique time points in the training instances. Only returned if `return_data` is True.

    """
    x = _validate_time_x(x, times)
    times = x[:, -1]
    states = x[:, :-1]
    unique_times = unique(times)
    n_times = len(unique_times)
    densities = []
    predictors = []

    for i, time in enumerate(unique_times):
        mask = times == time
        n_cells = arraysum(mask)
        logger.info(
            f"[{i+1} of {n_times}] Computing density for {n_cells:,} cells at time point {time}."
        )
        if n_cells < warn_below:
            logger.warning(
                f"Time point {time} only has {n_cells:,} cells. "
                "This could lead to inaccurate estimation of the time length scale `ls_time`."
            )

        x_at_time = x[mask, :-1]

        est = DensityEstimator(
            nn_distances=nn_distances[mask], **density_estimator_kwargs
        )
        est.fit(x_at_time)
        densities.append(est.predict(states))
        predictors.append(est)

    densities = stack(densities)
    corrs = corrcoef(densities)
    delta_t = abs(unique_times.reshape(-1, 1) - unique_times.reshape(1, -1)).reshape(
        -1, 1
    )

    def ls_loss(log_ls):
        ls = exp(log_ls)
        covs = cov_func_curry(ls)(delta_t, zeros((1, 1))).reshape((n_times, n_times))
        return norm(covs - corrs)

    opt = ScipyMinimize(fun=ls_loss, method="L-BFGS-B", jit=False).run(0.0)
    ls = exp(opt.params).item()

    if return_data:
        return ls, densities, predictors, unique_times

    return ls



from collections import namedtuple
from functools import partial
from jax import random, vmap
from jax.numpy import log, pi, exp, stack, arange, sort, mean, zeros_like
from jax.numpy import sum as arraysum
from jax.scipy.special import gammaln
import jax.scipy.stats.norm as norm
import jax
from jax.example_libraries.optimizers import adam
from jaxopt import ScipyMinimize
from .conditional import (
    FullConditionalMean,
    FullConditionalMeanTime,
    LandmarksConditionalMean,
    LandmarksConditionalMeanTime,
    LandmarksConditionalMeanCholesky,
    LandmarksConditionalMeanCholeskyTime,
)
from .util import ensure_2d, Exp, DEFAULT_JITTER


DEFAULT_N_ITER = 100
DEFAULT_INIT_LEARN_RATE = 1e-1
DEFAULT_NUM_SAMPLES = 40
DEFAULT_OPTIMIZER = "L-BFGS-B"
DEFAULT_JIT = False


def _normal(k):
    R"""
    Builds the log pdf of :math:`z \sim \text{Normal}(0, I)`.

    :param k: The size of :math:`z`.
    :type k: int
    :return: The log pdf of :math:`z`.
    :rtype: function
    """

    def logpdf(z):
        return -(1 / 2) * arraysum(z**2) - (k / 2) * log(2 * pi)

    return logpdf


def _multivariate(mu, L):
    R"""
    Builds the transformation function from
    :math:`z \sim \text{Normal}(0, I) \rightarrow f \sim \text{Normal}(\mu, K')`,
    where :math:`I` is the identity matrix
    and :math:`K \approx K' = L L^\top`.

    :param mu: The Gaussian process mean :math:`\mu`.
    :type mu: float
    :param L: A matrix such that :math:`L L^\top \approx K`.
    :type L: array-like
    :return: A function :math:`z \rightarrow f`.
    :rtype: function
    """

    def transform(z):
        return L.dot(z) + mu

    return transform


def _nearest_neighbors(r, d):
    """
    Returns the likelihood function of log densities :math:`p` given the observed
    distances :math:`L(p | r) = P(r | p)`, for number of dimensions :math:`d`.
    :param r: The observed nearest neighbor distances.
    :type r: array-like
    :param d: The local dimensionality.
    :type d: int
    :return: The likelihood function.
    :rtype: function
    """
    const = (d * log(pi) / 2) - gammaln(d / 2 + 1)
    V = log(r) * d + const
    Vdr = log(d) + ((d - 1) * log(r)) + const

    def logpdf(log_density):
        A = exp(log_density + V)
        B = log_density + Vdr
        return arraysum(B - A)

    return logpdf


def _poisson(distances):
    """
    Returns the likelihood function of dimensionality and density given the
    observed k nearest-neighbor distances
    .
    :param distances: The observed nearest neighbor distances.
    :type distances: array-like
    :return: The likelihood function.
    :rtype: function
    """
    k = distances.shape[1]
    counts = arange(1, k + 1)

    ldist = sort(distances, axis=-1)
    ldist = log(ldist) + log(pi) / 2

    def V(d):
        """
        Return the log-volume of the n-sphere for the raidus related values in ldist.
        """
        return d * ldist - gammaln(d / 2 + 1)

    def logpdf(dims, log_dens):
        pred = log_dens[:, None] + V(dims[:, None])
        logp = pred * counts[None, :] - exp(pred) - gammaln(counts)[None, :]
        return arraysum(logp)

    return logpdf


def compute_transform(mu, L):
    R"""
    Computes a function transform that maps :math:`z \sim
    \text{Normal}(0, I) \rightarrow f \sim \text{Normal}(\mu, K')`,
    where :math:`I` is the identity matrix and :math:`K \approx K' = L L^\top`,
    where :math:`K` is the covariance matrix.

    :param mu: The Gaussian process mean:math:`\mu`.
    :type mu: float
    :param L: A matrix such that :math:`L L^\top \approx K`, where :math:`K` is the
        covariance matrix.
    :type L: array-like
    :return: transform - The transform function :math:`z \rightarrow f`.
    """
    return _multivariate(mu, L)


def compute_dimensionality_transform(mu_dim, mu_dens, L):
    R"""
    Computes a function transform that maps :math:`z \sim
    \text{Normal}(0, I) \rightarrow \log(f) \sim \text{Normal}(\mu, K')`,
    where :math:`I` is the identity matrix and :math:`K \approx K' = L L^\top`,
    where :math:`K` is the covariance matrix.

    :param mu: The Gaussian process mean :math:`\mu`.
    :type mu: float
    :param L: A matrix such that :math:`L L^\top \approx K`, where :math:`K` is the
        covariance matrix.
    :type L: array-like
    :return: transform - The transform function :math:`z \rightarrow f`.
    """

    dim_transform = _multivariate(mu_dim, L)
    dens_transform = _multivariate(mu_dens, L)

    def transform(z):
        dims, dens = z[0, :], z[1, :]
        return exp(dim_transform(dims)), dens_transform(dens)

    return transform


def compute_loss_func(nn_distances, d, transform, k):
    R"""
    Computes the Bayesian loss function -(prior(:math:`z`) +
    likelihood(transform(:math:`z`))).

    :param nn_distances: The observed nearest neighbor distances.
    :type nn_distances: array-like
    :param d: The dimensionality of the data.
    :type d: int
    :param transform:
        Maps :math:`z \sim \text{Normal}(0, I) \rightarrow f \sim \text{Normal}(\mu, K')`,
        where :math:`I` is the identity matrix and :math:`K \approx K' = L L^\top`,
        where :math:`K` is the covariance matrix.
    :type transform: function
    :param k: dimension of transform input
    :type k: int
    :return: loss_func - The Bayesian loss function
    :rtype: function, function
    """
    prior = _normal(k)
    likelihood = _nearest_neighbors(nn_distances, d)

    def loss_func(z):
        return -(prior(z) + likelihood(transform(z)))

    return loss_func


def compute_dimensionality_loss_func(distances, transform, k):
    R"""
    Computes the Bayesian loss function -(prior(:math:`z`) +
    likelihood(transform(:math:`z`))) for dimensionality inference.

    :param distances: The observed k nearest neighbor distances.
    :type distances: array-like
    :param transform:
        Maps :math:`z \sim \text{Normal}(0, I) \rightarrow \log(f) \sim \text{Normal}(\mu, K')`,
        where :math:`I` is the identity matrix and :math:`K \approx K' = L L^\top`,
        where :math:`K` is the covariance matrix.
    :type transform: function
    :param k: dimension of transform input
    :type k: int
    :return: loss_func - The Bayesian loss function
    :rtype: function, function
    """
    prior = _normal(k)
    likelihood = _poisson(distances)

    def loss_func(z):
        dims, log_dens = transform(z)
        return -(prior(z) + likelihood(dims, log_dens))

    return loss_func


def minimize_adam(
    loss_func,
    initial_value,
    n_iter=DEFAULT_N_ITER,
    init_learn_rate=DEFAULT_INIT_LEARN_RATE,
    jit=DEFAULT_JIT,
):
    R"""
    Minimizes function with a starting guess of initial_value using
    adam and exponentially decaying learning rate.

    :param loss_func: The loss function to minimize.
    :type loss_func: function
    :param initial_value: The initial guess.
    :type initial_value: array-like
    :param n_iter: The number of optimization iterations. Defaults to 100.
    :type n_iter: integer
    :param init_learn_rate: The initial learn rate. Defaults to 1.
    :type init_learn_rate: float
    :return: Results - A named tuple containing pre_transformation, opt_state, losses: The optimized
        parameters, final state of the optimizer, and history of loss values,
    :rtype: array-like, array-like, Object
    """

    def learn_schedule(i):
        return exp(-1e-2 * i) * init_learn_rate

    opt_init, opt_update, get_params = adam(learn_schedule)
    opt_state = opt_init(initial_value)
    val_grad = jax.value_and_grad(loss_func)
    if jit:
        val_grad = jax.jit(val_grad)

    def step(step, opt_state):
        value, grads = val_grad(get_params(opt_state))
        opt_state = opt_update(step, grads, opt_state)
        return value, opt_state

    losses = list()
    for i in range(n_iter):
        value, opt_state = step(i, opt_state)
        losses.append(value.item())
    pre_transformation = get_params(opt_state)
    losses = stack(losses)

    Results = namedtuple("Results", "pre_transformation opt_state losses")
    results = Results(pre_transformation, opt_state, losses)
    return results


def minimize_lbfgsb(loss_func, initial_value, jit=DEFAULT_JIT):
    R"""
    Minimizes function with a starting guess of initial_value.

    :param loss_func: Loss function to minimize.
    :type loss_func: function
    :param initial_value: Initial guess.
    :type initial_value: array-like
    :return: Results - A named tuple containing pre_transformation, opt_state,
        loss: The optimized parameters, final state of the optimizer, and the
        final loss value,
    :rtype: array-like, array-like, Object
    """
    opt = ScipyMinimize(fun=loss_func, method="L-BFGS-B", jit=jit).run(initial_value)
    Results = namedtuple("Results", "pre_transformation opt_state loss")
    results = Results(opt.params, opt.state, opt.state.fun_val.item())
    return results


def compute_log_density_x(pre_transformation, transform):
    R"""
    Computes the log density at the training points.

    :param pre_transformation: :math:`z \sim \text{Normal}(0, I)`
    :type pre_transformation: array-like
    :param transform: A function
        :math:`z \sim \text{Normal}(0, I) \rightarrow f \sim \text{Normal}(\mu, K')`,
        where :math:`I` is the identity matrix and :math:`K \approx K' = L L^\top`,
        where :math:`K` is the covariance matrix.
    :type transform: function
    :return: log_density_x - The log density at the training points.
    """
    return transform(pre_transformation)


def compute_conditional_mean(
    x,
    landmarks,
    pre_transformation,
    y,
    mu,
    cov_func,
    sigma=0,
    jitter=DEFAULT_JITTER,
):
    R"""
    Builds the mean function of the Gaussian process, conditioned on the
    function values (e.g., log-density) on x. Returns an instance of
    mellon.Predictor that acts as a function, defined on the whole domain of x.

    Parameters
    ----------
    x : array-like
        The training instances.
    landmarks : array-like or None
        The landmark points for fast sparse computation.
        Landmarks can be None if not using landmark points.
    pre_transformation : array-like or None
        The pre-transformed latent function representation.
    y : array-like
        The function values at each point in x.
    mu : float
        The original Gaussian process mean :math:`\mu`.
    cov_func : function
        The Gaussian process covariance function.
    sigma : float, optional
        White noise variance, by default 0.
    jitter : float, optional
        A small amount to add to the diagonal for stability, by default 1e-6.

    Returns
    -------
    mellon.Predictor
        The conditioned Gaussian process mean function.
    """

    if landmarks is None:
        return FullConditionalMean(
            x,
            y,
            mu,
            cov_func,
            jitter=jitter,
        )
    elif (
        pre_transformation is not None
        and pre_transformation.shape[0] == landmarks.shape[0]
    ):
        landmarks = ensure_2d(landmarks)
        return LandmarksConditionalMeanCholesky(
            landmarks,
            pre_transformation,
            mu,
            cov_func,
            sigma=sigma,
            jitter=jitter,
        )
    else:
        landmarks = ensure_2d(landmarks)
        return LandmarksConditionalMean(
            x,
            landmarks,
            y,
            mu,
            cov_func,
            sigma=sigma,
            jitter=jitter,
        )


def compute_conditional_mean_times(
    x,
    landmarks,
    pre_transformation,
    y,
    mu,
    cov_func,
    sigma=0,
    jitter=DEFAULT_JITTER,
):
    R"""
    Builds the mean function of the Gaussian process, conditioned on the
    function values (e.g., log-density) on x, taking into account the associated
    times of each sample. Returns an instance of mellon.Predictor that acts as a
    function, defined on the whole domain of x, and can be evaluated at any given time.

    Parameters
    ----------
    x : array-like
        The training instances.
    landmarks : array-like or None
        The landmark points for fast sparse computation.
        Landmarks can be None if not using landmark points.
    pre_transformation : array-like or None
        The pre-transformed latent function representation.
    y : array-like
        The function values at each point in x.
    mu : float
        The original Gaussian process mean :math:`\mu`.
    cov_func : function
        The Gaussian process covariance function.
    sigma : float, optional
        White noise variance, by default 0.
    jitter : float, optional
        A small amount to add to the diagonal for stability, by default 1e-6.

    Returns
    -------
    mellon.Predictor
        The conditioned Gaussian process mean function that also accepts a 'times'
        argument to account for time-sensitive inferences.
    """

    if landmarks is None:
        return FullConditionalMeanTime(
            x,
            y,
            mu,
            cov_func,
            jitter=jitter,
        )
    elif (
        pre_transformation is not None
        and pre_transformation.shape[0] == landmarks.shape[0]
    ):
        landmarks = ensure_2d(landmarks)
        return LandmarksConditionalMeanCholeskyTime(
            landmarks,
            pre_transformation,
            mu,
            cov_func,
            sigma=sigma,
            jitter=jitter,
        )
    else:
        landmarks = ensure_2d(landmarks)
        return LandmarksConditionalMeanTime(
            x,
            landmarks,
            y,
            mu,
            cov_func,
            sigma=sigma,
            jitter=jitter,
        )


def compute_conditional_mean_explog(
    x,
    landmarks,
    y,
    mu,
    cov_func,
    sigma=0,
    jitter=DEFAULT_JITTER,
):
    R"""
    Builds the mean function of the Gaussian process, conditioned on the
    function exponential values (e.g., dimensionality) on x.
    Returns a function that is defined on the whole domain of x.

    :param x: The training instances.
    :type x: array-like
    :param landmarks: The landmark points for fast sparse computation.
        Landmarks can be None if not using landmark points.
    :type landmarks: array-like
    :param y: The function values at each point in x.
    :type y: array-like
    :param mu: The original Gaussian process mean :math:`\mu`.
    :type mu: float
    :param cov_func: The Gaussian process covariance function.
    :type cov_func: function
    :param sigma: White moise veriance. Defaults to 0.
    :type sigma: float
    :param jitter: A small amount to add to the diagonal for stability. Defaults to 1e-6.
    :type jitter: float
    :return: conditional_mean - The conditioned Gaussian process mean function.
    :rtype: function
    """
    if landmarks is None:
        return Exp(
            FullConditionalMean(
                x,
                log(y),
                mu,
                cov_func,
                jitter=jitter,
            )
        )
    else:
        landmarks = ensure_2d(landmarks)
        return Exp(
            LandmarksConditionalMean(
                x,
                landmarks,
                log(y),
                mu,
                cov_func,
                sigma=sigma,
                jitter=jitter,
            )
        )


def generate_gaussian_sample(rng, mean, log_std):
    """
    Generates a single sample from a multivariate Gaussian with diagonal covariance.

    :param rng: random number generator
    :param mean: mean of the Gaussian distribution
    :param log_std: logarithm of standard deviation of the Gaussian distribution
    :return: sample from the Gaussian distribution
    """
    return mean + exp(log_std) * random.normal(rng, mean.shape)


def calculate_gaussian_logpdf(x, mean, log_std):
    """
    Calculates the log probability density function of a multivariate Gaussian with diagonal covariance.

    :param x: value to evaluate the Gaussian on
    :param mean: mean of the Gaussian distribution
    :param log_std: logarithm of standard deviation of the Gaussian distribution
    :return: log pdf of the Gaussian distribution
    """
    return arraysum(vmap(norm.logpdf)(x, mean, exp(log_std)))


def calculate_elbo(logprob, rng, mean, log_std):
    """
    Calculates the single-sample Monte Carlo estimate of the variational lower bound (ELBO).

    :param logprob: log probability of the sample
    :param rng: random number generator
    :param mean: mean of the Gaussian distribution
    :param log_std: logarithm of standard deviation of the Gaussian distribution
    :return: ELBO estimate
    """
    sample = generate_gaussian_sample(rng, mean, log_std)
    return logprob(sample) - calculate_gaussian_logpdf(sample, mean, log_std)


def calculate_batch_elbo(logprob, rng, params, num_samples):
    """
    Calculates the average ELBO over a batch of random samples.

    :param logprob: log probability of the sample
    :param rng: random number generator key
    :param params: parameters of the Gaussian distribution
    :param num_samples: number of samples in the batch
    :return: batch ELBO
    """
    rngs = random.split(rng, num_samples)
    vectorized_elbo = vmap(partial(calculate_elbo, logprob), in_axes=(0, None, None))
    return mean(vectorized_elbo(rngs, *params))


def run_advi(
    loss_func,
    initial_parameters,
    n_iter=DEFAULT_N_ITER,
    init_learn_rate=DEFAULT_INIT_LEARN_RATE,
    nsamples=DEFAULT_NUM_SAMPLES,
    jit=DEFAULT_JIT,
):
    """
    Performs automatic differentiation variational inference (ADVI) to fit a
    Gaussian approximation to an intractable, unnormalized density.

    :param loss_func: function to calculate the loss
    :param initial_parameters: initial parameters for the optimization
    :param n_iter: number of iterations for the optimization (default: DEFAULT_N_ITER)
    :param init_learn_rate: The initial learn rate. Defaults to 1.
    :type init_learn_rate: float
    :return: parameters, standard deviations, and loss after the optimization
    :return: Results - A named tuple containing pre_transformation,
        pre_transform_std, losses: The optimized parameters, the optimized
        standard deviations, and a history of ELBO values.
    :rtype: array-like, array-like, Object
    """

    def negative_logprob(x):
        return -loss_func(x)

    def objective(params, t):
        rng = random.PRNGKey(t)
        return -calculate_batch_elbo(negative_logprob, rng, params, nsamples)

    def learn_schedule(i):
        return exp(-1e-2 * i) * init_learn_rate

    init_mean, init_std = initial_parameters, -10 * zeros_like(initial_parameters)
    opt_init, opt_update, get_params = adam(learn_schedule)
    opt_state = opt_init((init_mean, init_std))

    def update(i, opt_state):
        params = get_params(opt_state)
        elbo, gradient = jax.value_and_grad(objective)(params, i)
        return opt_update(i, gradient, opt_state), elbo

    if jit:
        update = jax.jit(update)

    elbos = list()
    for t in range(n_iter):
        opt_state, elbo = update(t, opt_state)
        elbos.append(elbo.item())

    params, stds = get_params(opt_state)

    Results = namedtuple("Results", "pre_transformation pre_transformation_std losses")
    return Results(params, stds, elbos)



from .inference import (
    compute_transform,
    compute_dimensionality_transform,
    compute_loss_func,
    compute_dimensionality_loss_func,
    minimize_adam,
    minimize_lbfgsb,
    compute_log_density_x,
    compute_conditional_mean,
    compute_conditional_mean_times,
    compute_conditional_mean_explog,
    generate_gaussian_sample,
    calculate_gaussian_logpdf,
    calculate_elbo,
    calculate_batch_elbo,
    run_advi,
)

__all__ = [
    "compute_transform",
    "compute_dimensionality_transform",
    "compute_loss_func",
    "compute_dimensionality_loss_func",
    "minimize_adam",
    "minimize_lbfgsb",
    "compute_log_density_x",
    "compute_conditional_mean",
    "compute_conditional_mean_times",
    "compute_conditional_mean_explog",
    "generate_gaussian_sample",
    "calculate_gaussian_logpdf",
    "calculate_elbo",
    "calculate_batch_elbo",
    "run_advi",
]



from .util import (
    mle,
    distance,
    test_rank,
    local_dimensionality,
    Log,
    set_jax_config,
)

__all__ = [
    "mle",
    "distance",
    "test_rank",
    "local_dimensionality",
    "Log",
    "set_jax_config",
]



from .density_estimator import DensityEstimator
from .function_estimator import FunctionEstimator
from .dimensionality_estimator import DimensionalityEstimator
from .time_sensitive_density_estimator import TimeSensitiveDensityEstimator

__all__ = [
    "DensityEstimator",
    "FunctionEstimator",
    "DimensionalityEstimator",
    "TimeSensitiveDensityEstimator",
]



from .parameters import (
    compute_d,
    compute_d_factal,
    compute_landmarks,
    compute_nn_distances,
    compute_nn_distances_within_time_points,
    compute_distances,
    compute_ls,
    compute_cov_func,
    compute_L,
    compute_mu,
    compute_initial_value,
    compute_initial_dimensionalities,
)
from .compute_ls_time import compute_ls_time

__all__ = [
    "compute_d",
    "compute_d_factal",
    "compute_landmarks",
    "compute_nn_distances",
    "compute_nn_distances_within_time_points",
    "compute_distances",
    "compute_ls",
    "compute_cov_func",
    "compute_L",
    "compute_mu",
    "compute_initial_value",
    "compute_initial_dimensionalities",
    "compute_ls_time",
]



from .decomposition import DEFAULT_RANK, DEFAULT_METHOD
from .base_model import BaseEstimator, DEFAULT_COV_FUNC
from .inference import (
    compute_dimensionality_transform,
    compute_dimensionality_loss_func,
    compute_log_density_x,
    compute_conditional_mean,
    compute_conditional_mean_explog,
    DEFAULT_N_ITER,
    DEFAULT_INIT_LEARN_RATE,
    DEFAULT_JIT,
    DEFAULT_OPTIMIZER,
)
from .parameters import (
    compute_distances,
    compute_mu,
    compute_initial_dimensionalities,
    DEFAULT_N_LANDMARKS,
)
from .util import (
    DEFAULT_JITTER,
    Log,
    local_dimensionality,
)
from .validation import (
    _validate_positive_int,
    _validate_float,
    _validate_array,
)


logger = Log()


class DimensionalityEstimator(BaseEstimator):
    R"""
    This class provides a non-parametric method for estimating local dimensionality and density.
    It uses Bayesian inference with a Gaussian process prior and a normal distribution for local scaling rates.
    The class caches all intermediate computations as instance variables,
    enabling users to view intermediate results and
    save computational time by passing precomputed values to a new model instance.

    Parameters
    ----------
    cov_func_curry: function or type, optional (default=Matern52)
        A generator for the Gaussian process covariance function. It should be
        a curry function taking one length scale argument
        and returning a covariance function of the form k(x, y) :math:`\rightarrow` float.

    n_landmarks: int, optional (default=5000)
        The number of landmark points. If less than 1 or greater than or equal
        to the number of training points,
        inducing points are not computed or used.

    rank: int or float, optional (default=0.99)
        The rank of the approximate covariance matrix. When interpreted as an
        integer, an :math:`n \times` rank matrix
        :math:`L` is computed such that :math:`L L^\top \approx K`, where
        :math:`K` is the exact :math:`n \times n` covariance matrix.
        When interpreted as a float (between 0.0 and 1.0), the rank/size of
        :math:`L` is chosen such that the included eigenvalues of the covariance
        between landmark points account for the specified percentage of the total eigenvalues.

    method: str, optional (default='auto')
        Determines whether the `rank` parameter is interpreted as a fixed
        number of eigenvectors ('fixed'), a percentage of eigenvalues ('percent'),
        or determined automatically ('auto'). In 'auto' mode, `rank` is treated
        as a fixed number if it is an integer, or a percentage if it's a float.

    jitter: float, optional (default=1e-6)
        A small amount added to the diagonal of the covariance matrix to ensure
        numerical stability by keeping eigenvalues away from 0.

    optimizer: str, optional (default='L-BFGS-B')
        The optimizer to use for maximum a posteriori density estimation.
        It can be either 'L-BFGS-B', 'adam', or 'advi'.

    n_iter: int, optional (default=100)
        The number of iterations for optimization.

    init_learn_rate: float, optional (default=1)
        The initial learning rate for the optimizer.

    landmarks: array-like or None, optional
        Points used to quantize the data for approximate covariance. If None,
        landmarks are set as k-means centroids
        with k=n_landmarks. If the number of landmarks is greater than or equal
        to the number of training points, this parameter is ignored.

    k: int, optional (default=10)
        The number of nearest neighbor distances to consider.

    distances: array-like or None, optional
        The k nearest neighbor distances at each data point. If None, these
        distances are computed automatically using KDTree (if data dimensionality is < 20)
        or BallTree otherwise.

    d: array-like, optional
        The estimated local intrinsic dimensionality of the data. This is only
        used to initialize the density estimation.
        If None, an empirical estimate is used.

    mu_dim: float or None, optional (default=0)
        The mean of the Gaussian process for log intrinsic dimensionality :math:`\mu_D`.

    mu_dens: float or None, optional
        The mean of the Gaussian process for log density :math:`\mu_\rho`. If None,
        `mu_dens` is set to the 1st percentile of :math:`\text{mle}(\text{nn_distances}, d) - 10`.

    ls : float or None, optional
        The length scale for the Gaussian process covariance function.
        If None (default), the length scale is automatically selected based on
        a heuristic link between the nearest neighbor distances and the optimal
        length scale.

    ls_factor : float, optional
        A scaling factor applied to the length scale when it's automatically
        selected. It is used to manually adjust the automatically chosen length
        scale for finer control over the model's sensitivity to variations in the data.

    cov_func: function or None, optional
        The Gaussian process covariance function of the form k(x, y)
        :math:`\rightarrow` float. If None, the covariance function is generated
        automatically as `cov_func = cov_func_curry(ls)`.

    L: array-like or None, optional
        A matrix such that :math:`L L^\top \approx K`, where :math:`K` is the
        covariance matrix. If None, `L` is computed automatically.

    initial_value: array-like or None, optional
        The initial guess for optimization. If None, an optimized :math:`z` is
        found that minimizes :math:`||Lz + \mu_\cdot - mle|| + ||z||`, where
        :math:`\text{mle}` is the maximum likelihood estimate
        for density initialization and the neighborhood-based local intrinsic
        dimensionality for dimensionality initialization.

    jit: bool, optional (default=False)
        If True, use JAX's just-in-time compilation for loss and its gradient during optimization.
    """

    def __init__(
        self,
        cov_func_curry=DEFAULT_COV_FUNC,
        n_landmarks=DEFAULT_N_LANDMARKS,
        rank=DEFAULT_RANK,
        method=DEFAULT_METHOD,
        jitter=DEFAULT_JITTER,
        optimizer=DEFAULT_OPTIMIZER,
        n_iter=DEFAULT_N_ITER,
        init_learn_rate=DEFAULT_INIT_LEARN_RATE,
        landmarks=None,
        k=10,
        distances=None,
        d=None,
        mu_dim=0,
        mu_dens=None,
        ls=None,
        ls_factor=1,
        cov_func=None,
        L=None,
        initial_value=None,
        jit=DEFAULT_JIT,
    ):
        super().__init__(
            cov_func_curry=cov_func_curry,
            n_landmarks=n_landmarks,
            rank=rank,
            method=method,
            jitter=jitter,
            optimizer=optimizer,
            n_iter=n_iter,
            init_learn_rate=init_learn_rate,
            landmarks=landmarks,
            nn_distances=None,
            d=d,
            mu=mu_dens,
            ls=ls,
            ls_factor=ls_factor,
            cov_func=cov_func,
            L=L,
            initial_value=initial_value,
            jit=jit,
        )
        self.k = _validate_positive_int(k, "k")
        self.mu_dim = _validate_float(mu_dim, "mu_dim")
        self.mu_dens = _validate_float(mu_dens, "mu_dens", optional=True)
        self.distances = _validate_array(distances, "distances", optional=True)
        self.transform = None
        self.loss_func = None
        self.opt_state = None
        self.losses = None
        self.pre_transformation = None
        self.local_dim_x = None
        self.log_density_x = None
        self.local_dim_func = None
        self.log_density_func = None

    def __repr__(self):
        name = self.__class__.__name__
        string = (
            f"{name}("
            f"cov_func_curry={self.cov_func_curry}, "
            f"n_landmarks={self.n_landmarks}, "
            f"rank={self.rank}, "
            f"method='{self.method}', "
            f"jitter={self.jitter}, "
            f"optimizer='{self.optimizer}', "
            f"n_iter={self.n_iter}, "
            f"init_learn_rate={self.init_learn_rate}, "
            f"landmarks={self.landmarks}, "
        )
        if self.distances is None:
            string += "distances=None, "
        else:
            string += "distances=distances, "
        string += (
            f"d={self.d}, "
            f"mu_dim={self.mu_dim}, "
            f"mu_dens={self.mu_dens}, "
            f"ls={self.ls}, "
            f"cov_func={self.cov_func}, "
        )
        if self.L is None:
            string += "L=None, "
        else:
            string += "L=L, "
        if self.initial_value is None:
            string += "initial_value=None, "
        else:
            string += "initial_value=initial_value, "
        string += f"jit={self.jit}" ")"
        return string

    def _compute_mu_dens(self):
        nn_distances = self.nn_distances
        d = self.d
        mu = compute_mu(nn_distances, d)
        return mu

    def _compute_d(self):
        x = self.x
        d = local_dimensionality(x, neighbor_idx=None)
        return d

    def _compute_initial_value(self):
        x = self.x
        d = self.d
        nn_distances = self.nn_distances
        mu_dim = self.mu_dim
        mu_dens = self.mu_dens
        L = self.L
        initial_value = compute_initial_dimensionalities(
            x, mu_dim, mu_dens, L, nn_distances, d
        )
        return initial_value

    def _compute_transform(self):
        mu_dim = self.mu_dim
        mu_dens = self.mu_dens
        L = self.L
        transform = compute_dimensionality_transform(mu_dim, mu_dens, L)
        return transform

    def _compute_distances(self):
        x = self.x
        k = self.k
        logger.info("Computing distances.")
        distances = compute_distances(x, k=k)
        return distances

    def _compute_nn_distances(self):
        distances = self.distances
        return distances[:, 0]

    def _compute_loss_func(self):
        distances = self.distances
        transform = self.transform
        k = self.initial_value.shape[0]
        loss_func = compute_dimensionality_loss_func(distances, transform, k)
        return loss_func

    def _set_local_dim_x(self):
        pre_transformation = self.pre_transformation
        transform = self.transform
        local_dim_x, log_density_x = compute_log_density_x(
            pre_transformation, transform
        )
        self.local_dim_x = local_dim_x
        self.log_density_x = log_density_x

    def _set_local_dim_func(self):
        x = self.x
        landmarks = self.landmarks
        local_dim_x = self.local_dim_x
        mu = self.mu_dim
        cov_func = self.cov_func
        jitter = self.jitter
        logger.info("Computing predictive dimensionality function.")
        log_dim_func = compute_conditional_mean_explog(
            x,
            landmarks,
            local_dim_x,
            mu,
            cov_func,
            jitter=jitter,
        )
        self.local_dim_func = log_dim_func

    def _set_log_density_func(self):
        x = self.x
        landmarks = self.landmarks
        pre_transformation = self.pre_transformation
        log_density_x = self.log_density_x
        mu = self.mu_dens
        cov_func = self.cov_func
        jitter = self.jitter
        logger.info("Computing predictive density function.")
        log_density_func = compute_conditional_mean(
            x,
            landmarks,
            pre_transformation,
            log_density_x,
            mu,
            cov_func,
            jitter=jitter,
        )
        self.log_density_func = log_density_func

    def prepare_inference(self, x):
        R"""
        Set all attributes in preparation for optimization, but do not
        perform Bayesian inference. It is not necessary to call this
        function before calling fit.

        :param x: The training instances to estimate density function.
        :type x: array-like
        :return: loss_func, initial_value - The Bayesian loss function and
            initial guess for optimization.
        :rtype: function, array-like
        """
        if x is None:
            x = self.x
            if self.x is None:
                message = "Required argument x is missing and self.x has not been set."
                raise ValueError(message)
        else:
            if self.x is not None and self.x is not x:
                message = (
                    "self.x has been set already, but is not equal to the argument x."
                )
                raise ValueError(message)

        x = self.set_x(x)
        self._prepare_attribute("distances")
        self._prepare_attribute("nn_distances")
        self._prepare_attribute("d")
        self._prepare_attribute("mu_dens")
        self._prepare_attribute("ls")
        self._prepare_attribute("cov_func")
        self._prepare_attribute("landmarks")
        self._prepare_attribute("L")
        self._prepare_attribute("initial_value")
        self._prepare_attribute("transform")
        self._prepare_attribute("loss_func")
        return self.loss_func, self.initial_value

    def run_inference(self, loss_func=None, initial_value=None, optimizer=None):
        R"""
        Perform Bayesian inference, optimizing the pre_transformation parameters.
        If you would like to run your own inference procedure, use the loss_function
        and initial_value attributes and set pre_transformation to the optimized
        parameters.

        :param loss_func: The Bayesian loss function. If None, uses the stored
            loss_func attribute.
        :type loss_func: function
        :param initial_value: The initial guess for optimization. If None, uses
            the stored initial_value attribute.
        :type initial_value: array-like
        :return: pre_transformation - The optimized parameters.
        :rtype: array-like
        """
        if loss_func is not None:
            self.loss_func = loss_func
        if initial_value is not None:
            self.initial_value = initial_value
        if optimizer is not None:
            self.optimizer = optimizer
        self._run_inference()
        return self.pre_transformation

    def process_inference(self, pre_transformation=None, build_predict=True):
        R"""
        Use the optimized parameters to compute the local dimensionality at the
        training points. If build_predict, also build the prediction function.

        :param pre_transformation: The optimized parameters. If None, uses the stored
            pre_transformation attribute.
        :type pre_transformation: array-like
        :param build_predict: Whether or not to build the prediction function.
            Defaults to True.
        :type build_predict: bool
        :return: local_dim_x - The local dimensionality
        :rtype: array-like
        """
        if pre_transformation is not None:
            self.pre_transformation = pre_transformation
        self._set_local_dim_x()
        if build_predict:
            self._set_local_dim_func()
            self._set_log_density_func()
        return self.local_dim_x, self.log_density_x

    def fit(self, x=None, build_predict=True):
        """
        Trains the model from start to finish. This includes preparing for inference, running inference,
        and processing the inference results.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features), default=None
            The training instances where `n_samples` is the number of samples and `n_features`
            is the number of features.

        build_predict : bool, default=True
            Whether or not to construct the prediction function after training.

        Returns
        -------
        self
            A fitted instance of this estimator.
        """

        self.prepare_inference(x)
        self.run_inference()
        self.process_inference(build_predict=build_predict)
        return self

    @property
    def predict_density(self):
        """
        Predicts the log density with an adaptive unit for each data point in `x`.
        The unit of density depends on the dimensionality of the data.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features)
            The new data for which to predict the log density.

        Returns
        -------
        array-like
            The predicted log density for each test point in `x`.

        Example
        -------

        >>> log_density = model.predict_density(Xnew)

        """
        if self.log_density_func is None:
            self._set_log_density_func()
        return self.log_density_func

    @property
    def predict(self):
        """
        Returns an instance of the :class:`mellon.Predictor` class, which predicts the dimensionality
        at each point in `x`.

        This instance includes a __call__ method, which can be used to predict the dimensionality.
        The instance also supports serialization features, allowing for saving and loading the predictor's
        state. For more details, refer to :class:`mellon.Predictor`.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features)
            The new data for which to predict the dimensionality.

        Returns
        -------
        array-like
            The predicted dimensionality for each test point in `x`.

        Example
        -------

        >>> log_density = model.predict(Xnew)

        """
        if self.local_dim_func is None:
            self._set_local_dim_func()
        return self.local_dim_func

    def fit_predict(self, x=None, build_predict=False):
        """
        Trains the model using the provided training data, and then makes predictions
        on the trained data points. This function performs Bayesian inference to compute
        the local dimensionality and returns the computed local dimensionality at each
        training point.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features), default=None
            The training instances to estimate the local dimensionality function, where
            `n_samples` is the number of samples and `n_features` is the number of features.
            Each sample is an array of features representing a point in the feature space.

        build_predict : bool, default=False
            Whether or not to build the prediction function after training.

        Returns
        -------
        array-like of shape (n_samples,)
            The local dimensionality at each training point in `x`.

        Raises
        ------
        ValueError
            If the argument `x` does not match `self.x` which was already set in a previous operation.
        """
        if self.x is not None and self.x is not x:
            message = "self.x has been set already, but is not equal to the argument x."
            error = ValueError(message)
            logger.error(error)
            raise error
        if self.x is None and x is None:
            message = "Required argument x is missing and self.x has not been set."
            error = ValueError(message)
            logger.error(error)
            raise error
        if x is None:
            x = self.x
        else:
            x = _validate_array(x, "x")

        self.fit(x, build_predict=build_predict)
        return self.local_dim_x



from jax.numpy import sqrt, exp, square
from .base_cov import Covariance
from .util import distance, select_active_dims


class Matern32(Covariance):
    R"""
    Implementation of the Matern-3/2 kernel function, a member of the Matern
    family of kernels.

    The Matern-3/2 kernel function is defined as:

    .. math::

        (1 + \frac{\sqrt{3}||x-y||}{l}) \cdot e^{-\frac{\sqrt{3}||x-y||}{l}}

    where `x` and `y` are input vectors and `l` is the length-scale.

    This class can be used as a function curry, meaning it can be called
    like a function on two inputs `x` and `y`.

    Parameters
    ----------
    ls : float, optional
        The length-scale parameter, which controls the width of the kernel.
        Larger values result in wider kernels, and smaller values in narrower kernels.
        Default is 1.0.

    active_dims : array-like, slice or scalar, optional
        The indices of the active dimensions. If specified, the kernel function
        will only be computed over these dimensions. Default is None, which means
        all dimensions are active.
    """

    def __init__(self, ls=1.0, active_dims=None):
        super().__init__()
        self.ls = ls
        self.active_dims = active_dims

    def k(self, x, y):
        R"""
        Compute the Matern-3/2 kernel function between inputs `x` and `y`.

        The kernel function is computed over the active dimensions, specified
        by the `active_dims` parameter during initialization.

        Parameters
        ----------
        x : array-like
            First input array.
        y : array-like
            Second input array.

        Returns
        -------
        similarity : float
            The computed kernel function.
        """
        x = select_active_dims(x, self.active_dims)
        y = select_active_dims(y, self.active_dims)
        r = sqrt(3.0) * distance(x, y) / self.ls
        similarity = (r + 1) * exp(-r)
        return similarity


class Matern52(Covariance):
    R"""
    Implementation of the Matern-5/2 kernel function, a member of the Matern
    family of kernels.

    The Matern-5/2 kernel function is defined as:

    .. math::

        (1 + \frac{\sqrt{5}||x-y||}{l} + \frac{5||x-y||^2}{3l^2})
        \cdot e^{-\frac{\sqrt{5}||x-y||}{l}}

    where `x` and `y` are input vectors and `l` is the length-scale.

    This class can be used as a function curry, meaning it can be called
    like a function on two inputs `x` and `y`.

    Parameters
    ----------
    ls : float, optional
        The length-scale parameter, which controls the width of the kernel.
        Larger values result in wider kernels, and smaller values in narrower kernels.
        Default is 1.0.

    active_dims : array-like, slice or scalar, optional
        The indices of the active dimensions. If specified, the kernel function
        will only be computed over these dimensions. Default is None, which means
        all dimensions are active.
    """

    def __init__(self, ls=1.0, active_dims=None):
        super().__init__()
        self.ls = ls
        self.active_dims = active_dims

    def k(self, x, y):
        R"""
        Compute the Matern-5/2 kernel function between inputs `x` and `y`.

        The kernel function is computed over the active dimensions, specified
        by the `active_dims` parameter during initialization.

        Parameters
        ----------
        x : array-like
            First input array.
        y : array-like
            Second input array.

        Returns
        -------
        similarity : float
            The computed kernel function.
        """
        x = select_active_dims(x, self.active_dims)
        y = select_active_dims(y, self.active_dims)
        r = sqrt(5.0) * distance(x, y) / self.ls
        similarity = (r + square(r) / 3 + 1) * exp(-r)
        return similarity


class ExpQuad(Covariance):
    R"""
    Exponentiated Quadratic kernel, also known as the squared exponential or the Gaussian kernel.

    The kernel is defined as:

    .. math::

        e^{-\frac{||x-y||^2}{2 l^2}}

    This class can be used as a function curry, meaning it can be called
    like a function on two inputs `x` and `y`.

    Parameters
    ----------
    ls : float, optional
        The length-scale parameter, which controls the width of the kernel.
        Larger values result in wider kernels, and smaller values in narrower kernels.
        Default is 1.0.

    active_dims : array-like, slice or scalar, optional
        The indices of the active dimensions. If specified, the kernel function
        will only be computed over these dimensions. Default is None, which means
        all dimensions are active.
    """

    def __init__(self, ls=1.0, active_dims=None):
        super().__init__()
        self.ls = ls
        self.active_dims = active_dims

    def k(self, x, y):
        R"""
        Compute the Exponentiated Quadratic kernel function between inputs `x` and `y`.

        The kernel function is computed over the active dimensions, specified
        by the `active_dims` parameter during initialization.

        Parameters
        ----------
        x : array-like
            First input array.
        y : array-like
            Second input array.

        Returns
        -------
        similarity : float
            The computed kernel function.
        """
        x = select_active_dims(x, self.active_dims)
        y = select_active_dims(y, self.active_dims)
        r = distance(x, y) / self.ls
        similarity = exp(-square(r) / 2)
        return similarity


class Exponential(Covariance):
    R"""
    Exponential kernel.

    The kernel is defined as:

    .. math::

        e^{-\frac{||x-y||}{2l}}

    This class can be used as a function curry, meaning it can be called
    like a function on two inputs `x` and `y`.

    Parameters
    ----------
    ls : float, optional
        The length-scale parameter, which controls the width of the kernel.
        Larger values result in wider kernels, and smaller values in narrower kernels.
        Default is 1.0.

    active_dims : array-like, slice or scalar, optional
        The indices of the active dimensions. If specified, the kernel function
        will only be computed over these dimensions. Default is None, which means
        all dimensions are active.
    """

    def __init__(self, ls=1.0, active_dims=None):
        super().__init__()
        self.ls = ls
        self.active_dims = active_dims

    def k(self, x, y):
        R"""
        Compute the Exponential kernel function between inputs `x` and `y`.

        The kernel function is computed over the active dimensions, specified
        by the `active_dims` parameter during initialization.

        Parameters
        ----------
        x : array-like
            First input array.
        y : array-like
            Second input array.

        Returns
        -------
        similarity : float
            The computed kernel function.
        """
        x = select_active_dims(x, self.active_dims)
        y = select_active_dims(y, self.active_dims)
        r = distance(x, y) / self.ls
        similarity = exp(-r / 2)
        return similarity


class RatQuad(Covariance):
    R"""
    Rational Quadratic kernel.

    The kernel is defined as:

    .. math::

        (1 + \frac{||x-y||^2}{2 \alpha l^2})^{-\alpha l}

    This class can be used as a function curry, meaning it can be called
    like a function on two inputs `x` and `y`.

    Parameters
    ----------
    ls : float, optional
        The length-scale parameter, which controls the width of the kernel.
        Larger values result in wider kernels, and smaller values in narrower kernels.
        Default is 1.0.

    alpha : float
        The alpha parameter of the Rational Quadratic kernel.

    active_dims : array-like, slice or scalar, optional
        The indices of the active dimensions. If specified, the kernel function
        will only be computed over these dimensions. Default is None, which means
        all dimensions are active.
    """

    def __init__(self, alpha, ls=1.0, active_dims=None):
        super().__init__()
        self.ls = ls
        self.alpha = alpha
        self.active_dims = active_dims

    def k(self, x, y):
        R"""
        Compute the Rational Quadratic kernel function between inputs `x` and `y`.

        The kernel function is computed over the active dimensions, specified
        by the `active_dims` parameter during initialization.

        Parameters
        ----------
        x : array-like
            First input array.
        y : array-like
            Second input array.

        Returns
        -------
        similarity : float
            The computed kernel function.
        """
        x = select_active_dims(x, self.active_dims)
        y = select_active_dims(y, self.active_dims)
        r = distance(x, y) / self.ls
        similarity = (square(r) / (2 * self.alpha) + 1) ** -self.alpha
        return similarity



from .decomposition import DEFAULT_RANK, DEFAULT_METHOD
from .base_model import BaseEstimator, DEFAULT_COV_FUNC
from .inference import (
    compute_transform,
    compute_loss_func,
    compute_log_density_x,
    compute_conditional_mean_times,
    DEFAULT_N_ITER,
    DEFAULT_INIT_LEARN_RATE,
    DEFAULT_JIT,
    DEFAULT_OPTIMIZER,
)
from .parameters import (
    compute_nn_distances_within_time_points,
    compute_landmarks_rescale_time,
    compute_cov_func,
    compute_d,
    compute_d_factal,
    compute_mu,
    compute_initial_value,
    DEFAULT_N_LANDMARKS,
)
from .compute_ls_time import compute_ls_time
from .util import (
    DEFAULT_JITTER,
    Log,
)
from .validation import (
    _validate_time_x,
    _validate_positive_float,
    _validate_string,
    _validate_array,
    _validate_bool,
)


DEFAULT_D_METHOD = "embedding"

logger = Log()


class TimeSensitiveDensityEstimator(BaseEstimator):
    R"""
    A class for non-parametric density estimation with time sensitivity.
    It performs Bayesian inference with
    a Gaussian process prior and Nearest Neighbor likelihood. All intermediate computations
    are cached as instance variables, which allows viewing intermediate results and
    saving computation time by passing precomputed values as arguments to a new model.

    Parameters
    ----------
    cov_func_curry : function or type
        The generator of the Gaussian process covariance function.
        This must be a curry that takes one length scale argument and returns a
        covariance function of the form k(x, y) :math:`\rightarrow` float.
        Defaults to Matern52.

    n_landmarks : int
        The number of landmark points. If less than 1 or greater than or equal to the
        number of training points, inducing points will not be computed or used.
        Defaults to 5000.

    rank : int or float
        The rank of the approximate covariance matrix. If rank is an int, an :math:`n \times`
        rank matrix :math:`L` is computed such that :math:`L L^\top \approx K`, where `K` is the
        exact :math:`n \times n` covariance matrix. If rank is a float 0.0 :math:`\le` rank
        :math:`\le` 1.0, the rank/size of :math:`L` is selected such that the included eigenvalues
        of the covariance between landmark points account for the specified percentage of the sum
        of eigenvalues. Defaults to 0.99.

    method : str
        Determines how the rank is interpreted: as a fixed number of eigenvectors ('fixed'), a
        percent of eigenvalues ('percent'), or automatically ('auto'). If 'auto', the rank is
        interpreted as a fixed number of eigenvectors if it is an int and as a percent of
        eigenvalues if it is a float. This parameter is provided for clarity in the ambiguous case
        of 1 vs 1.0. Defaults to 'auto'.

    d_method : str
        The method to compute the intrinsic dimensionality of the data. Implemented options are
         - 'embedding': uses the embedding dimension `x.shape[1]`
         - 'fractal': uses the average fractal dimension (experimental)

        Defaults to 'embedding'.

    jitter : float
        A small amount added to the diagonal of the covariance matrix to bind eigenvalues
        numerically away from 0, ensuring numerical stability. Defaults to 1e-6.

    optimizer : str
        The optimizer for the maximum a posteriori or posterior density estimation. Options are
        'L-BFGS-B', stochastic optimizer 'adam', or automatic differentiation variational
        inference 'advi'. Defaults to 'L-BFGS-B'.

    n_iter : int
        The number of optimization iterations. Defaults to 100.

    init_learn_rate : float
        The initial learning rate. Defaults to 1.

    landmarks : array-like or None
        The points used to quantize the data for the approximate covariance. If None,
        landmarks are set as k-means centroids with k=n_landmarks. This is ignored if n_landmarks
        is greater than or equal to the number of training points. Defaults to None.

    nn_distances : array-like or None
        The nearest neighbor distances at each data point within each time point.
        If None, the nearest neighbor
        distances are computed automatically, using a KDTree if the dimensionality of the data
        is less than 20, or a BallTree otherwise. Defaults to None.

    normalize_per_time_point : bool, optional
        If True, the computation of nearest neighbor distances incorporates a normalization step
        based on the number of samples within each time point group. This process mitigates potential
        bias in the density estimation due to unequal sample distribution across different time points.
        Note that if `nn_distance` is provided, this parameter is ignored as the provided distances
        are used directly.
        Defaults to False.

    d : int, array-like or None
        The intrinsic dimensionality of the data, i.e., the dimensionality of the embedded
        manifold. If None, `d` is set to the size of axis 1 of the training data points.
        Defaults to None.

    mu : float or None
        The mean :math:`\mu` of the Gaussian process. If None, sets :math:`\mu` to the 1st
        percentile of :math:`\text{mle}(\text{nn_distances}, d) - 10`, where :math:`\text{mle} =
        \log(\text{gamma}(d/2 + 1)) - (d/2) \cdot \log(\pi) - d \cdot \log(\text{nn_distances})`.
        Defaults to None.

    ls : float or None, optional
        The length scale for the Gaussian process covariance function.
        If None (default), the length scale is automatically selected based on
        a heuristic link between the nearest neighbor distances and the optimal
        length scale.

    ls_time : float or None
        The length scale of the Gaussian process covariance function for the time dimension.
        If None, `ls_time` is set to the length scale that best induces a covariance
        (using the `cov_func_curry`) between the time points that best mimics the
        Pearson correlation observed between densities of the individual time points. If `cov_func`
        is supplied explicitly, `ls_time` has no effect. Defaults to None.

    ls_factor : float, optional
        A scaling factor applied to the length scale when it's automatically
        selected. It is used to manually adjust the automatically chosen length

    ls_time_factor : float, optional
        A scaling factor applied to the time length scale (`ls_time`) when it's automatically
        selected. This allows for manual adjustment of the automatically determined time length scale.
        Defaults to 1.

    density_estimator_kwargs : dict, optional
        A dictionary of keyword arguments to be passed for timepoint-specific
        density estimation during the automatic selection of the time length scale `ls_time`.
        This parameter allows custom configuration for the density estimation process.
        Note that this parameter has no effect if `ls_time` is specified explicitly.
        Default is an empty dictionary ({}).

    cov_func : mellon.Covariance or None
        The Gaussian process covariance function of the form k(x, y) :math:`\rightarrow` float.
        Should be an instance of a class that inherits from :class:`mellon.Covariance`.
        If None, the covariance function `cov_func` is automatically generated as
        `cov_func_curry(ls, active_dims=slice(None, -1)) * cov_func_curry(ls_time, active_dims=-1)`.
        Defaults to None.

    L : array-like or None
        A matrix such that :math:`L L^\top \approx K`, where :math:`K` is the covariance matrix.
        If None, `L` is computed automatically. Defaults to None.

    initial_value : array-like or None
        The initial guess for optimization. If None, the value :math:`z` that minimizes
        :math:`||Lz + \mu - mle|| + ||z||` is found, where :math:`\text{mle} = \log(\text{gamma}(d/2 + 1))
        - (d/2) \cdot \log(\pi) - d \cdot \log(\text{nn_distances})` and :math:`d` is the intrinsic
        dimensionality of the data. Defaults to None.

    _save_intermediate_ls_times : bool
        Determines whether the intermediate results obtained during the computation of `ls_time` are retained for debugging.
        When set to True, the results will be stored in `self.densities`, `self.predictors`, and `self.numeric_stages`.
        Defaults to False.


    jit : bool
        Use jax just-in-time compilation for loss and its gradient during optimization.
        Defaults to False.
    """

    def __init__(
        self,
        cov_func_curry=DEFAULT_COV_FUNC,
        n_landmarks=DEFAULT_N_LANDMARKS,
        rank=DEFAULT_RANK,
        method=DEFAULT_METHOD,
        d_method=DEFAULT_D_METHOD,
        jitter=DEFAULT_JITTER,
        optimizer=DEFAULT_OPTIMIZER,
        n_iter=DEFAULT_N_ITER,
        init_learn_rate=DEFAULT_INIT_LEARN_RATE,
        landmarks=None,
        nn_distances=None,
        normalize_per_time_point=False,
        d=None,
        mu=None,
        ls=None,
        ls_time=None,
        ls_factor=1,
        ls_factor_times=1,
        density_estimator_kwargs=dict(),
        cov_func=None,
        L=None,
        initial_value=None,
        _save_intermediate_ls_times=False,
        jit=DEFAULT_JIT,
    ):
        super().__init__(
            cov_func_curry=cov_func_curry,
            n_landmarks=n_landmarks,
            rank=rank,
            jitter=jitter,
            optimizer=optimizer,
            n_iter=n_iter,
            init_learn_rate=init_learn_rate,
            landmarks=landmarks,
            nn_distances=nn_distances,
            d=d,
            mu=mu,
            ls=ls,
            ls_factor=ls_factor,
            cov_func=cov_func,
            L=L,
            initial_value=initial_value,
            jit=jit,
        )
        if not isinstance(density_estimator_kwargs, dict):
            raise ValueError("density_estimator_kwargs needs to be a dictionary.")
        self.density_estimator_kwargs = density_estimator_kwargs
        self.d_method = _validate_string(
            d_method, "d_method", choices={"fractal", "embedding"}
        )
        self.ls_time = _validate_positive_float(ls_time, "ls_time", optional=True)
        self.ls_factor_times = _validate_positive_float(
            ls_factor_times, "ls_factor_times"
        )
        self._save_intermediate_ls_times = _save_intermediate_ls_times
        self.normalize_per_time_point = _validate_bool(
            normalize_per_time_point, "normalize_per_time_point"
        )
        self.transform = None
        self.loss_func = None
        self.opt_state = None
        self.losses = None
        self.pre_transformation = None
        self.log_density_x = None
        self.log_density_func = None

    def __repr__(self):
        name = self.__class__.__name__
        string = (
            f"{name}("
            f"cov_func_curry={self.cov_func_curry}, "
            f"n_landmarks={self.n_landmarks}, "
            f"rank={self.rank}, "
            f"method='{self.method}', "
            f"jitter={self.jitter}, "
            f"optimizer='{self.optimizer}', "
            f"n_iter={self.n_iter}, "
            f"init_learn_rate={self.init_learn_rate}, "
            f"landmarks={self.landmarks}, "
        )
        if self.nn_distances is None:
            string += "nn_distances=None, "
        else:
            string += "nn_distances=nn_distances, "
        string += (
            f"d={self.d}, "
            f"mu={self.mu}, "
            f"ls={self.ls}, "
            f"ls_time={self.ls_time}, "
            f"cov_func={self.cov_func}, "
        )
        if self.L is None:
            string += "L=None, "
        else:
            string += "L=L, "
        if self.initial_value is None:
            string += "initial_value=None, "
        else:
            string += "initial_value=initial_value, "
        string += f"jit={self.jit}" ")"
        return string

    def _compute_d(self):
        x = self.x[:, :-1]
        if self.d_method == "fractal":
            logger.warning("Using EXPERIMENTAL fractal dimensionality selection.")
            d = compute_d_factal(x)
            logger.info(f"Using d={d}.")
        else:
            d = compute_d(x)
        if d > 50:
            message = f"""The detected dimensionality of the data is over 50,
            which is likely to cause numerical instability issues.
            Consider running a dimensionality reduction algorithm, or
            if this number of dimensions is intended, explicitly pass
            d={self.d} as a parameter."""
            raise ValueError(message)
        return d

    def _compute_mu(self):
        nn_distances = self.nn_distances
        d = self.d
        mu = compute_mu(nn_distances, d)
        return mu

    def _compute_initial_value(self):
        nn_distances = self.nn_distances
        d = self.d
        mu = self.mu
        L = self.L
        initial_value = compute_initial_value(nn_distances, d, mu, L)
        return initial_value

    def _compute_transform(self):
        mu = self.mu
        L = self.L
        transform = compute_transform(mu, L)
        return transform

    def _compute_loss_func(self):
        nn_distances = self.nn_distances
        d = self.d
        transform = self.transform
        k = self.initial_value.shape[0]
        loss_func = compute_loss_func(nn_distances, d, transform, k)
        return loss_func

    def _compute_nn_distances(self):
        x = self.x
        normalize_per_time_point = self.normalize_per_time_point
        logger.info("Computing nearest neighbor distances within time points.")
        nn_distances = compute_nn_distances_within_time_points(
            x,
            normalize=normalize_per_time_point,
        )
        return nn_distances

    def _compute_ls_time(self):
        nn_distances = self.nn_distances
        x = self.x
        cov_func_curry = self.cov_func_curry
        _save_intermediate_ls_times = self._save_intermediate_ls_times
        density_estimator_kwargs = {
            "cov_func_curry": self.cov_func_curry,
            "d_method": self.d_method,
            "d": self.d,
            "optimizer": self.optimizer,
            "ls": self.ls,
            "ls_factor": self.ls_factor,
            "jit": self.jit,
            "mu": self.mu,
        }
        density_estimator_kwargs.update(self.density_estimator_kwargs)
        logger.info(
            "Initiating density computation for each time point to estimate the 'ls_time' parameter. "
            "You can directly specify 'ls_time' to bypass this computation-intensive step."
        )
        ls = compute_ls_time(
            nn_distances,
            x,
            cov_func_curry,
            return_data=_save_intermediate_ls_times,
            density_estimator_kwargs=density_estimator_kwargs,
        )
        if _save_intermediate_ls_times:
            logger.info(
                "Storing `self.densities`, `self.predictors`, and `self.numeric_stages`."
            )
            ls, self.densities, self.predictors, self.numeric_stages = ls
        ls *= self.ls_factor_times
        return ls

    def _compute_landmarks(self):
        x = self.x
        ls = self.ls
        ls_time = self.ls_time
        n_landmarks = self.n_landmarks
        n_samples = x.shape[0]
        if n_samples > 100 * n_landmarks and n_samples > 1e6:
            logger.info(
                f"Large number of {n_samples:,} cells and "
                f"small number of {n_landmarks:,} landmarks. Consider "
                "computing k-means on a subset of cells and passing "
                "the results as 'landmarks' to speed up the process."
            )
        landmarks = compute_landmarks_rescale_time(
            x, ls, ls_time, n_landmarks=n_landmarks
        )
        return landmarks

    def _compute_cov_func(self):
        cov_func_curry = self.cov_func_curry
        ls = self.ls
        ls_time = self.ls_time
        cov_func = compute_cov_func(cov_func_curry, ls, ls_time)
        logger.info("Using covariance function %s.", str(cov_func))
        return cov_func

    def _set_log_density_x(self):
        pre_transformation = self.pre_transformation
        transform = self.transform
        log_density_x = compute_log_density_x(pre_transformation, transform)
        self.log_density_x = log_density_x

    def _set_log_density_func(self):
        x = self.x
        landmarks = self.landmarks
        pre_transformation = self.pre_transformation
        log_density_x = self.log_density_x
        mu = self.mu
        cov_func = self.cov_func
        jitter = self.jitter
        logger.info("Computing predictive function.")
        log_density_func = compute_conditional_mean_times(
            x,
            landmarks,
            pre_transformation,
            log_density_x,
            mu,
            cov_func,
            jitter=jitter,
        )
        self.log_density_func = log_density_func

    def prepare_inference(self, x, times=None):
        R"""
        Prepares for optimization without performing Bayesian inference.
        This method sets all attributes required for optimization.
        It is not required to call this method manually before calling `fit`.

        Parameters
        ----------
        x : array-like
            The training instances for which the density function will be estimated.
            If 'times' is None, the last column of 'x' is interpreted as the times.
            Shape must be (n_samples, n_features).

        times : array-like, optional
            An array encoding the time points associated with each cell/row in 'x'.
            If provided, it overrides the last column of 'x' as the times.
            Shape must be either (n_samples,) or (n_samples, 1).

        Returns
        -------
        loss_func : function
            The Bayesian loss function that will be minimized during optimization.

        initial_value : array-like
            The initial guess for the optimization process.
        """

        if x is None:
            x = self.x
            if self.x is None:
                message = "Required argument x is missing and self.x has not been set."
                raise ValueError(message)
        else:
            x = _validate_time_x(x, times)
            if self.x is not None and self.x is not x:
                message = (
                    "self.x has been set already, but is not equal to the argument x."
                )
                raise ValueError(message)

        x = self.set_x(x)
        self._prepare_attribute("nn_distances")
        self._prepare_attribute("d")
        self._prepare_attribute("mu")
        self._prepare_attribute("ls")
        self._prepare_attribute("ls_time")
        self._prepare_attribute("cov_func")
        self._prepare_attribute("landmarks")
        self._prepare_attribute("L")
        self._prepare_attribute("initial_value")
        self._prepare_attribute("transform")
        self._prepare_attribute("loss_func")
        return self.loss_func, self.initial_value

    def run_inference(self, loss_func=None, initial_value=None, optimizer=None):
        R"""
        Perform Bayesian inference, optimizing the pre_transformation parameters.
        If you would like to run your own inference procedure, use the loss_function
        and initial_value attributes and set pre_transformation to the optimized
        parameters.

        :param loss_func: The Bayesian loss function. If None, uses the stored
            loss_func attribute.
        :type loss_func: function
        :param initial_value: The initial guess for optimization. If None, uses
            the stored initial_value attribute.
        :type initial_value: array-like
        :return: pre_transformation - The optimized parameters.
        :rtype: array-like
        """
        if loss_func is not None:
            self.loss_func = loss_func
        if initial_value is not None:
            self.initial_value = initial_value
        if optimizer is not None:
            self.optimizer = optimizer
        self._run_inference()
        return self.pre_transformation

    def process_inference(self, pre_transformation=None, build_predict=True):
        R"""
        Use the optimized parameters to compute the log density at the
        training points. If build_predict, also build the prediction function.

        :param pre_transformation: The optimized parameters. If None, uses the stored
            pre_transformation attribute.
        :type pre_transformation: array-like
        :param build_predict: Whether or not to build the prediction function.
            Defaults to True.
        :type build_predict: bool
        :return: log_density_x - The log density
        :rtype: array-like
        """
        if pre_transformation is not None:
            self.pre_transformation = _validate_array(
                pre_transformation, "pre_transformation"
            )
        self._set_log_density_x()
        if build_predict:
            self._set_log_density_func()
        return self.log_density_x

    def fit(self, x=None, times=None, build_predict=True):
        R"""
        Fit the model from end to end.

        Parameters
        ----------
        x : array-like, optional
            The training instances to estimate density function.
            If 'x' is not provided and 'self.x' is also None, a ValueError is raised.
        times : array-like, optional
            An array encoding the time points associated with each cell/row in 'x'.
            Shape must be either (n_samples,) or (n_samples, 1).
        build_predict : bool, optional
            Whether or not to build the prediction function. Defaults to True.

        Returns
        -------
        self : object
            A fitted instance of this estimator.

        Raises
        ------
        ValueError
            If both 'x' and 'self.x' are None or if 'x' is provided and not equal to 'self.x'.
        """

        self.prepare_inference(x, times)
        self.run_inference()
        self.process_inference(build_predict=build_predict)
        return self

    @property
    def predict(self):
        R"""
        An instance of the :class:`mellon.Predictor` that predicts the log density at each point in x.

        The instance contains a __call__ method which can be used to predict the log density.
        This instance also supports serialization features which allows for saving
        and loading the predictor state. Refer to mellon.Predictor documentation for more details.

        Note that the last column of the input array `x` should contain the time information.

        Parameters
        ----------
        x : array-like
            The new data to predict, where the last column should contain the time information.

        Returns
        -------
        log_density : array-like
            The log density at each test point in `x`.

        Example
        -------

        >>> log_density = model.predict(Xnew)

        """
        if self.log_density_func is None:
            self._set_log_density_func()
        return self.log_density_func

    def fit_predict(self, x=None, times=None, build_predict=False):
        R"""
        Perform Bayesian inference and return the log density at training points.

        :param x: The training instances to estimate density function.
        :type x: array-like
        :return: log_density_x - The log density at each training point in x.
        """
        if x is not None:
            x = _validate_time_x(x, times)
        if self.x is not None and self.x is not x:
            message = "self.x has been set already, but is not equal to the argument x."
            error = ValueError(message)
            logger.error(error)
            raise error
        if self.x is None and x is None:
            message = "Required argument x is missing and self.x has not been set."
            error = ValueError(message)
            logger.error(error)
            raise error
        if x is None:
            x = self.x

        self.fit(x, build_predict=build_predict)
        return self.log_density_x



from collections.abc import Iterable

from jax.numpy import asarray, concatenate, isscalar, full, ndarray
from jax.errors import ConcretizationTypeError


def _validate_time_x(x, times=None, n_features=None, cast_scalar=False):
    """
    Validates and concatenates 'x' and 'times' if 'times' is provided.

    If 'times' is provided, the function reshapes it if necessary and checks for
    matching number of samples in 'x' and 'times', before concatenating.

    If 'n_features' is provided, the function checks if 'x' has the correct number
    of features.

    Parameters
    ----------
    x : array-like
        The training instances for which the density function will be estimated.
        Shape must be (n_samples, n_features).

    times : array-like, optional
        An array encoding the time points associated with each cell/row in 'x'.
        Shape must be either (n_samples,) or (n_samples, 1).

    n_features : int, optional
        The expected number of features in 'x' including 'times' if it is provided.

    cast_scalar : bool, optional
        If true and 'times' is a scalar, it will be cast to a 1D array with a length
        equal to the number of samples in 'x'.

    Returns
    -------
    array-like
        The concatenated array of 'x' and 'times' (if provided), or 'x' if 'times'
        is not provided.

    Raises
    ------
    ValueError
        If 'times' is not a 1D or 2D array with 1 column, or 'x' and 'times' don't
        have the same number of samples, or 'x' does not have the expected number
        of features.
    """

    x = _validate_array(x, "x", ndim=2)
    if (
        cast_scalar
        and times is not None
        and (isscalar(times) or all(s == 1 for s in times.shape))
    ):
        times = full(x.shape[0], times)
    times = _validate_array(times, "times", optional=True, ndim=(1, 2))

    if times is not None:
        # Validate 'times' shape
        if times.ndim == 1:
            times = times.reshape(-1, 1)
        elif times.ndim != 2 or times.shape[1] != 1:
            raise ValueError("'times' must be a 1D array or a 2D array with 1 column.")

        # Check that 'x' and 'times' have the same number of samples
        if x.shape[0] != times.shape[0]:
            raise ValueError(
                "'x' and 'times' must have the same number of samples. "
                f"Got {x.shape[0]} for 'x' and {times.shape[0]} for 'times'."
            )

        # Concatenate 'x' and 'times'
        x = concatenate((x, times), axis=1)

    if n_features is not None:
        if x.shape[1] == n_features - 1 and times is None:
            raise ValueError(
                f"Expected {n_features} features including 'times' in 'x' but "
                f"only found {x.shape[1]} features and 'times' is not provided."
            )
        elif x.shape[1] != n_features:
            raise ValueError(
                f"Wrong number of features in 'x'. Expected {n_features} but got {x.shape[1]}."
            )

    return x


def _validate_float_or_int(value, param_name, optional=False):
    if value is None and optional:
        return None

    if not isinstance(value, (float, int)):
        raise ValueError(f"'{param_name}' should be a positive integer or float number")
    return value


def _validate_positive_float(value, param_name, optional=False):
    if value is None and optional:
        return None

    if not isinstance(value, (float, int)) or value < 0:
        raise ValueError(f"'{param_name}' should be a positive float number")
    return float(value)


def _validate_float(value, param_name, optional=False):
    """
    Validates if the input is a float or can be converted to a float.

    Parameters
    ----------
    value : object
        Input to be checked.
    param_name : str
        Name of the input parameter, used for error messaging.
    optional : bool, optional
        If True and the input is None, None will be returned. Otherwise, if the input is None,
        it raises an error. By default, False.

    Returns
    -------
    float
        The input converted to float.

    Raises
    ------
    ValueError
        If the input is not a float, integer, or a one-element array, or if the input is None and
        optional=False.
    """
    if value is None:
        if optional:
            return None
        else:
            raise ValueError(
                f"'{param_name}' is None, but is required to be a float number"
            )

    if isinstance(value, ndarray) and value.size == 1:
        try:
            value = value.item()
        except ConcretizationTypeError:
            # this must be a JAX tracer
            return value

    if not isinstance(value, (float, int)):
        raise ValueError(f"'{param_name}' should be a float number")

    return float(value)


def _validate_positive_int(value, param_name, optional=False):
    if optional and value is None:
        return None
    if not isinstance(value, int) or value < 0:
        raise ValueError(f"'{param_name}' should be a positive integer number")
    return value


def _validate_array(iterable, name, optional=False, ndim=None):
    """
    Validates and converts an iterable to a numpy array of type float.
    Allows Jax's JVPTracer objects and avoids explicit conversion in these cases.

    Parameters
    ----------
    iterable : iterable or None
        The iterable to be validated and converted. If optional=True, can also be None.

    name : str
        The name of the variable, used in error messages.

    optional : bool, optional
        If True, 'iterable' can be None and the function will return None in this case.
        If False and 'iterable' is None, a TypeError is raised. Defaults to False.

    ndim : int or tuple of ints, optional
        The number of dimensions the array must have. If specified and the number of dimensions
        of 'iterable' is not in 'ndim', a ValueError is raised.

    Returns
    -------
    numpy.ndarray or jax._src.interpreters.ad.JVPTracer
        The input iterable converted to a numpy array of type float, or the input JVPTracer.

    Raises
    ------
    TypeError
        If 'iterable' is not iterable and not None, or if 'iterable' is None and optional=False.

    ValueError
        If 'iterable' can't be converted to a numpy array of type float,
        or if the number of dimensions of 'iterable' is not in 'ndim'.
    """

    if iterable is None:
        if optional:
            return None
        else:
            raise TypeError(f"'{name}' can't be None.")

    if isinstance(iterable, Iterable):
        array = asarray(iterable, dtype=float)
    else:
        raise TypeError(f"'{name}' should be iterable, got {type(iterable)} instead.")

    if ndim is not None:
        if isinstance(ndim, int):
            ndim = (ndim,)
        if array.ndim not in ndim:
            raise ValueError(
                f"'{name}' must be a {ndim}-dimensional array, got {array.ndim}-dimensional array instead."
            )

    return array


def _validate_bool(value, name):
    if not isinstance(value, bool):
        raise TypeError(f"{name} should be of type bool, got {type(value)} instead.")

    return value


def _validate_string(value, name, choices=None):
    if not isinstance(value, str):
        raise TypeError(f"{name} should be of type str, got {type(value)} instead.")

    if choices and value not in choices:
        raise ValueError(f"{name} should be one of {choices}, got '{value}' instead.")

    return value


def _validate_float_or_iterable_numerical(value, name, optional=False):
    if value is None and optional:
        return None

    if isinstance(value, (int, float)):
        return float(value)

    if isinstance(value, Iterable):
        try:
            return asarray(value, dtype=float)
        except Exception:
            raise ValueError(f"Could not convert {name} to a numeric array.")

    raise TypeError(
        f"{name} should be of type int, float or iterable, got {type(value)} instead."
    )


def _validate_cov_func_curry(cov_func_curry, cov_func, param_name):
    if cov_func_curry is None and cov_func is None:
        raise ValueError(
            "At least one of 'cov_func_curry' and 'cov_func' must not be None"
        )

    from .base_cov import Covariance

    if cov_func_curry is not None:
        if not isinstance(cov_func_curry, type) or not issubclass(
            cov_func_curry, Covariance
        ):
            raise ValueError(f"'{param_name}' must be a subclass of mellon.Covariance")
    return cov_func_curry


def _validate_cov_func(cov_func, param_name, optional=False):
    if cov_func is None and optional:
        return None
    from .base_cov import Covariance

    if not isinstance(cov_func, Covariance):
        raise ValueError(
            f"'{param_name}' must be an instance of a subclass of mellon.Covariance"
        )
    return cov_func


def _validate_1d(x):
    """
    Validates that `x` can be cast to a JAX array with exactly 1 dimension and float data type.

    Parameters
    ----------
    x : array-like or scalar
        The input data to be validated and cast.

    Returns
    -------
    array-like
        The validated and cast data. If `x` is a scalar, it is transformed into a 1D array with a single element.

    Raises
    ------
    ValueError
        If `x` cannot be cast to a JAX array with exactly 1 dimension.
    """
    x = asarray(x, dtype=float)

    # Add an extra dimension if x is a scalar
    if x.ndim == 0:
        x = x[None]

    if x.ndim != 1:
        raise ValueError("`x` must be exactly 1-dimensional.")

    return x



from jax.numpy import dot, square, isnan, any
from jax.numpy.linalg import cholesky
from jax.scipy.linalg import solve_triangular
from .util import ensure_2d, make_serializable, stabilize, DEFAULT_JITTER, Log
from .base_predictor import Predictor, PredictorTime


logger = Log()


class _FullConditionalMean:
    def __init__(
        self,
        x,
        y,
        mu,
        cov_func,
        sigma=0,
        jitter=DEFAULT_JITTER,
    ):
        """
        The mean function of the conditioned Gaussian process.

        :param x: The training instances.
        :type x: array-like
        :param y: The function value at each point in x.
        :type y: array-like
        :param mu: The original Gaussian process mean.
        :type mu: float
        :param cov_func: The Gaussian process covariance function.
        :type cov_func: function
        :param sigma: White moise standard deviation. Defaults to 0.
        :type sigma: float
        :param jitter: A small amount to add to the diagonal for stability. Defaults to 1e-6.
        :type jitter: float
        :return: conditional_mean - The conditioned Gaussian process mean function.
        :rtype: function
        """
        x = ensure_2d(x)
        sigma2 = square(sigma)
        K = cov_func(x, x)
        sigma2 = max(sigma2, jitter)
        L = cholesky(stabilize(K, jitter=sigma2))
        if any(isnan(L)):
            message = (
                f"Covariance not positively definite with jitter={jitter}. "
                "Consider increasing the jitter for numerical stabilization."
            )
            logger.error(message)
            raise ValueError(message)
        r = y - mu
        weights = solve_triangular(L.T, solve_triangular(L, r, lower=True))

        self.cov_func = cov_func
        self.x = x
        self.weights = weights
        self.mu = mu
        self.n_input_features = x.shape[1]

    def _data_dict(self):
        return {
            "x": make_serializable(self.x),
            "weights": make_serializable(self.weights),
            "mu": make_serializable(self.mu),
        }

    def _predict(self, Xnew):
        cov_func = self.cov_func
        x = self.x
        weights = self.weights
        mu = self.mu

        Kus = cov_func(Xnew, x)
        return mu + dot(Kus, weights)


class FullConditionalMean(_FullConditionalMean, Predictor):
    pass


class FullConditionalMeanTime(_FullConditionalMean, PredictorTime):
    pass


class _LandmarksConditionalMean:
    def __init__(
        self,
        x,
        xu,
        y,
        mu,
        cov_func,
        sigma=0,
        jitter=DEFAULT_JITTER,
    ):
        """
        The mean function of the conditioned low rank gp, where rank
        is less than the number of landmark points.

        :param x: The training instances.
        :type x: array-like
        :param xu: The landmark points.
        :type xu: array-like
        :param y: The function value at each point in x.
        :type y: array-like
        :param mu: The original Gaussian process mean.
        :type mu: float
        :param cov_func: The Gaussian process covariance function.
        :type cov_func: function
        :param sigma: White moise standard deviation. Defaults to 0.
        :type sigma: float
        :param jitter: A small amount to add to the diagonal for stability. Defaults to 1e-6.
        :type jitter: float
        :return: conditional_mean - The conditioned Gaussian process mean function.
        :rtype: function
        """
        x = ensure_2d(x)
        xu = ensure_2d(xu)
        sigma2 = square(sigma)
        Kuu = cov_func(xu, xu)
        Kuf = cov_func(xu, x)
        Luu = cholesky(stabilize(Kuu, jitter))
        if any(isnan(Luu)):
            message = (
                f"Covariance of landmarks not positively definite with jitter={jitter}. "
                "Consider increasing the jitter for numerical stabilization."
            )
            logger.error(message)
            raise ValueError(message)
        A = solve_triangular(Luu, Kuf, lower=True)
        sigma2 = max(sigma2, jitter)
        L_B = cholesky(stabilize(dot(A, A.T), sigma2))
        r = y - mu
        c = solve_triangular(L_B, dot(A, r), lower=True)
        z = solve_triangular(L_B.T, c)
        weights = solve_triangular(Luu.T, z)

        self.cov_func = cov_func
        self.landmarks = xu
        self.weights = weights
        self.mu = mu
        self.n_input_features = xu.shape[1]

    def _data_dict(self):
        return {
            "landmarks": make_serializable(self.landmarks),
            "weights": make_serializable(self.weights),
            "mu": make_serializable(self.mu),
        }

    def _predict(self, Xnew):
        cov_func = self.cov_func
        xu = self.landmarks
        weights = self.weights
        mu = self.mu

        Kus = cov_func(Xnew, xu)
        return mu + dot(Kus, weights)


class LandmarksConditionalMean(_LandmarksConditionalMean, Predictor):
    pass


class LandmarksConditionalMeanTime(_LandmarksConditionalMean, PredictorTime):
    pass


class _LandmarksConditionalMeanCholesky:
    def __init__(
        self,
        xu,
        pre_transformation,
        mu,
        cov_func,
        sigma=0,
        jitter=DEFAULT_JITTER,
    ):
        """
        The mean function of the conditioned low rank gp, where rank
        is the number of landmark points.

        :param xu: The landmark points.
        :type xu: array-like
        :param pre_transformation: The pre transform latent function representation.
        :type pre_transformation: array-like
        :param mu: The original Gaussian process mean.
        :type mu: float
        :param cov_func: The Gaussian process covariance function.
        :type cov_func: function
        :param sigma: White moise standard deviation. Defaults to 0.
        :type sigma: float
        :param jitter: A small amount to add to the diagonal for stability. Defaults to 1e-6.
        :type jitter: float
        :return: conditional_mean - The conditioned Gaussian process mean function.
        :rtype: function
        """
        xu = ensure_2d(xu)
        sigma2 = square(sigma)
        K = cov_func(xu, xu)
        sigma2 = max(sigma2, jitter)
        L = cholesky(stabilize(K, jitter=sigma2))
        if any(isnan(L)):
            message = (
                f"Covariance not positively definite with jitter={jitter}. "
                "Consider increasing the jitter for numerical stabilization."
            )
            logger.error(message)
            raise ValueError(message)
        weights = solve_triangular(L.T, pre_transformation)

        self.cov_func = cov_func
        self.landmarks = xu
        self.weights = weights
        self.mu = mu
        self.n_input_features = xu.shape[1]

    def _data_dict(self):
        return {
            "landmarks": make_serializable(self.landmarks),
            "weights": make_serializable(self.weights),
            "mu": make_serializable(self.mu),
        }

    def _predict(self, Xnew):
        cov_func = self.cov_func
        xu = self.landmarks
        weights = self.weights
        mu = self.mu

        Kus = cov_func(Xnew, xu)
        return mu + dot(Kus, weights)


class LandmarksConditionalMeanCholesky(
    _LandmarksConditionalMeanCholesky, PredictorTime
):
    pass


class LandmarksConditionalMeanCholeskyTime(
    _LandmarksConditionalMeanCholesky, PredictorTime
):
    pass



from jax.numpy import exp, log, quantile, stack, unique, empty
from jax.numpy import sum as arraysum
from jax import random
from sklearn.cluster import k_means
from sklearn.linear_model import Ridge
from sklearn.neighbors import BallTree, KDTree
from .util import mle, local_dimensionality, Log, ensure_2d, DEFAULT_JITTER
from .decomposition import (
    _check_method,
    _full_rank,
    _full_decomposition_low_rank,
    _standard_low_rank,
    _modified_low_rank,
    DEFAULT_RANK,
    DEFAULT_METHOD,
)
from .validation import _validate_time_x, _validate_positive_float


DEFAULT_N_LANDMARKS = 5000

logger = Log()


def compute_landmarks(x, n_landmarks=DEFAULT_N_LANDMARKS):
    R"""
    Computes the landmark points as k-means centroids.

    Landmark points are used to approximate the underlying structure of the
    input space. If the number of landmarks (`n_landmarks`) is zero or
    exceeds the number of available data points, the function will return None.

    Parameters
    ----------
    x : array-like
        The input data for which landmarks should be computed.
        Shape must be (n_samples, n_features).
    n_landmarks : int, optional
        The desired number of landmark points. If less than 2 or greater
        than the number of data points, the function will return None.
        Defaults to DEFAULT_N_LANDMARKS.

    Returns
    -------
    landmark_points : array-like or None
        The coordinates of the computed landmark points, represented as
        k-means centroids. If no landmarks are computed, the function
        returns None. Shape is (n_landmarks, n_features).

    """
    if n_landmarks == 0:
        return None
    n = x.shape[0]
    x = ensure_2d(x)
    assert n_landmarks > 1, "n_landmarks musst be larger 1 or euqual to 0"
    if n_landmarks >= n:
        return None
    logger.info(f"Computing {n_landmarks:,} landmarks with k-means clustering.")
    return k_means(x, n_landmarks, n_init=1)[0]


def compute_landmarks_rescale_time(
    x, ls, ls_time, times=None, n_landmarks=DEFAULT_N_LANDMARKS
):
    R"""
    Computes landmark points for time-rescaled input data using k-means centroids.

    This function first rescales the temporal dimension of the input data by
    a factor derived from the spatial and temporal length scales (`ls` and `ls_time`).
    It then computes landmark points from the rescaled data. The last dimension
    of the landmarks is re-scaled back to the original time scale before being returned.

    Parameters
    ----------
    x : array-like
        The input data for which landmarks should be computed. If 'times' is
        None, the last column of 'x' is interpreted as the times.
        Shape must be (n_samples, n_features).
    ls : float
        Length scale of the spatial covariance kernel. Must be positive.
    ls_time : float
        Length scale of the temporal covariance kernel. Must be positive.
    times : array-like, optional
        An array encoding the time points associated with each sample in 'x'.
        If provided, it overrides the last column of 'x' as the times.
        Shape must be either (n_samples,) or (n_samples, 1).
    n_landmarks : int, optional
        The desired number of landmark points. Defaults to DEFAULT_N_LANDMARKS.

    Returns
    -------
    landmark_points : array-like or None
        The coordinates of the computed landmark points, represented as
        k-means centroids in the original space, including the re-scaled
        temporal dimension. If no landmarks are computed, the function
        returns None. Shape is (n_landmarks, n_features).

    """
    if n_landmarks == 0:
        return None

    ls = _validate_positive_float(ls, "ls")
    ls_time = _validate_positive_float(ls_time, "ls_time")
    x = _validate_time_x(x, times)
    time_factor = ls / ls_time
    x = x.at[:, -1].set(x[:, -1] * time_factor)
    landmarks = compute_landmarks(x, n_landmarks=n_landmarks)
    if landmarks is not None:
        try:
            landmarks = landmarks.at[:, -1].set(landmarks[:, -1] / time_factor)
        except AttributeError:
            # landmarks is not a jax array
            landmarks[:, -1] = landmarks[:, -1] / time_factor
    return landmarks


def compute_distances(x, k):
    R"""
    Computes the distance to the k nearest neighbor for each training instance.

    :param x: The training instances.
    :type x: array-like
    :param k: The number of nearest neighbors to consider.
    :return: distances - The k observed nearest neighbor distances.
    :rtype: array-like
    """
    x = ensure_2d(x)
    if x.shape[1] >= 20:
        tree = BallTree(x, metric="euclidean")
    else:
        tree = KDTree(x, metric="euclidean")
    distances = tree.query(x, k=k + 1)[0][:, 1:]
    return distances


def compute_nn_distances(x):
    """
    Compute the distance to the nearest neighbor for each instance in the provided training dataset.

    This function calculates the Euclidean distance between each instance in the dataset and its closest neighbor.
    It returns an array of these distances, ordered in the same way as the input instances.

    Parameters
    ----------
    x : array-like of shape (n_samples, n_features)
        An array-like object representing the training instances.

    Returns
    -------
    nn_distances : array-like of shape (n_samples,)
        An array of the Euclidean distances from each instance to its nearest neighbor in
        the input dataset. The ordering of the distances in this array corresponds to the
        ordering of the instances in the input data.

    """
    return compute_distances(x, 1)[:, 0]


def compute_nn_distances_within_time_points(x, times=None, normalize=False):
    R"""
    Computes the distance to the nearest neighbor for each training instance
    within the same time point group. It retains the original order of instances in `x`.

    Parameters
    ----------
    x : array-like
        The training instances.
        If 'times' is None, the last column of 'x' is interpreted as the times.
        Shape must be (n_samples, n_features).

    times : array-like, optional
        An array encoding the time points associated with each cell/row in 'x'.
        If provided, it overrides the last column of 'x' as the times.
        Shape must be either (n_samples,) or (n_samples, 1).

    normalize : bool, optional
        If True, distances are normalized by the number of samples within the same time point group.
        This normalization reduces potential bias in the density estimation arising from uneven
        sampling across different time points. Defaults to False.

    Returns
    -------
    nn_distances : array-like
        The observed nearest neighbor distances within the same time point group,
        preserving the order of instances in `x`.

    """
    x = _validate_time_x(x, times)
    unique_times = unique(x[:, -1])
    nn_distances = empty(x.shape[0])
    n_cells = x.shape[0]
    av_cells_per_tp = n_cells / len(unique_times)

    for time in unique_times:
        mask = x[:, -1] == time
        n_samples = arraysum(mask)
        if n_samples < 2:
            raise ValueError(
                f"Insufficient data: Only {n_samples} sample(s) found at time point {time}. "
                "Nearest neighbors cannot be computed with less than two samples per time point. "
                "Please confirm if you have provided the correct time axis. "
                "If the time points indeed have very few samples, consider aggregating nearby time points for better results, "
                "or you may specify `nn_distances` manually."
            )
        x_at_time = x[mask, :-1]
        nn_distances_at_time = compute_nn_distances(x_at_time)
        if normalize:
            nn_distances_at_time = nn_distances_at_time * n_samples / av_cells_per_tp
        nn_distances = nn_distances.at[mask].set(nn_distances_at_time)

    return nn_distances


def compute_d(x):
    R"""
    Computes the dimensionality of the data equal to the size of axis 1.

    :param x: The training instances.
    :type x: array-like
    """
    if len(x.shape) < 2:
        return 1
    return x.shape[1]


def compute_d_factal(x, k=10, n=500, seed=432):
    R"""
    Computes the dimensionality of the data based on the average fractal
    dimension around n randomly selected cells.

    :param x: The training instances.
    :type x: array-like
    :param n: Number of samples.
    :type n: int
    :param seed: Random seed for sampling.
    :type seed: int
    """
    if len(x.shape) < 2:
        return 1
    if n < x.shape[0]:
        key = random.PRNGKey(seed)
        idx = random.choice(key, x.shape[0], shape=(n,), replace=False)
        x_query = x[idx, ...]
    else:
        x_query = x
    local_dims = local_dimensionality(x, k=k, x_query=x_query)
    return local_dims.mean()


def compute_mu(nn_distances, d):
    R"""
    Computes mu equal to the 1th percentile of :math:`mle(nn\text{_}distances, d) - 10`,
    where :math:`mle =
    \log(\text{gamma}(d/2 + 1)) - (d/2) \cdot \log(\pi) - d \cdot \log(nn\text{_}distances)`

    :param nn_distances: The observed nearest neighbor distances.
    :type nn_distances: array-like
    :type d: The local dimensionality of the data.
    :type d: int
    :return: mu - The 1th percentile of :math:`mle(nn\text{_}distances, d) - 10`.
    :rtype: float
    """
    return quantile(mle(nn_distances, d), 0.01).item() - 10


def compute_ls(nn_distances):
    R"""
    Computes ls equal to the geometric mean of the nearest neighbor distances times a constant.

    :param nn_distances: The observed nearest neighbor distances.
    :type nn_distances: array-like
    :return: ls - The geometric mean of the nearest neighbor distances times a constant.
    :rtype: float
    """
    return exp(log(nn_distances).mean() + 3.0).item()


def compute_cov_func(cov_func_curry, ls, ls_time=None):
    R"""
    Computes the Gaussian process covariance function from its generator and length scales.

    Parameters
    ----------
    cov_func_curry : function or type
        The covariance function generator.

    ls : float
        The length scale of the covariance function.

    ls_time : float, optional
        The time-specific length scale of the covariance function.
        If provided, the returned covariance function will account for the time-specific
        dimension of the input data (last dimension assumed to be time). Defaults to None.

    Returns
    -------
    cov_func : mellon.Covariance instance
        The resulting Gaussian process covariance function. If `ls_time` is
        provided, the covariance function is a product of two covariance functions,
        one for the feature dimensions and one for the time dimension.
        Otherwise, it's a single covariance function considering only the feature dimensions.
    """
    if ls_time is not None:
        return cov_func_curry(ls, active_dims=slice(None, -1)) * cov_func_curry(
            ls_time, active_dims=-1
        )
    return cov_func_curry(ls)


def compute_L(
    x,
    cov_func,
    landmarks=None,
    rank=DEFAULT_RANK,
    method=DEFAULT_METHOD,
    jitter=DEFAULT_JITTER,
):
    R"""
    Compute an :math:`L` such that :math:`L L^\top \approx K`, where
    :math:`K` is the covariance matrix.

    :param x: The training instances.
    :type x: array-like
    :param cov_func: The Gaussian process covariance function.
    :type cov_func: function
    :param landmarks: The landmark points. If None, computes a full rank decompostion.
        Defaults to None.
    :type landmarks: array-like
    :param rank: The rank of the approximate covariance matrix.
        If rank is an int, an :math:`n \times` rank matrix
        :math:`L` is computed such that :math:`L L^\top \approx K`, the exact
        :math:`n \times n` covariance matrix.
        If rank is a float 0.0 :math:`\le` rank :math:`\le` 1.0, the rank/size
        of :math:`L` is selected such that the included eigenvalues of the covariance
        between landmark points account for the specified percentage of the
        sum of eigenvalues. Defaults to 0.999.
    :type rank: int or float
    :param method: Explicitly specifies whether rank is to be interpreted as a
        fixed number of eigenvectors or a percent of eigenvalues to include
        in the low rank approximation. Supports 'fixed', 'percent', or 'auto'.
        If 'auto', interprets rank as a fixed number of eigenvectors if it is
        an int and interprets rank as a percent of eigenvalues if it is a float.
        Defaults to 'auto'.
    :type method: str
    :param jitter: A small amount to add to the diagonal. Defaults to 1e-6.
    :type jitter: float
    :return: :math:`L` - A matrix such that :math:`L L^\top \approx K`.
    :rtype: array-like
    """
    x = ensure_2d(x)
    n_samples = x.shape[0]
    if landmarks is None:
        n = x.shape[0]
        method = _check_method(rank, n, method)

        if type(rank) is int and rank == n or type(rank) is float and rank == 1.0:
            logger.info(
                f"Doing full-rank Cholesky decomposition for {n_samples:,} samples."
            )
            return _full_rank(x, cov_func, jitter=jitter)
        else:
            logger.info(
                f"Doing full-rank singular value decomposition for {n_samples:,} samples."
            )
            return _full_decomposition_low_rank(
                x, cov_func, rank=rank, method=method, jitter=jitter
            )
    else:
        landmarks = ensure_2d(landmarks)

        n_landmarks = landmarks.shape[0]
        method = _check_method(rank, n_landmarks, method)

        if (
            type(rank) is int
            and rank == n_landmarks
            or type(rank) is float
            and rank == 1.0
        ):
            logger.info(
                "Doing low-rank Cholesky decomposition for "
                f"{n_samples:,} samples and {n_landmarks:,} landmarks."
            )
            return _standard_low_rank(x, cov_func, landmarks, jitter=jitter)
        else:
            logger.info(
                "Doing low-rank improved Nyström decomposition for "
                f"{n_samples:,} samples and {n_landmarks:,} landmarks."
            )
            return _modified_low_rank(
                x, cov_func, landmarks, rank=rank, method=method, jitter=jitter
            )


def compute_initial_value(nn_distances, d, mu, L):
    R"""
    Computes the initial value for Maximum A Posteriori optimization with Ridge regression,
    such that the initial value :math:`z` minimizes
    :math:`||Lz + mu - mle(nn\text{_}distances, d)|| + ||z||`.

    :param nn_distances: The observed nearest neighbor distances.
    :type nn_distances: array-like
    :param d: The local dimensionality of the data.
    :type d: int
    :param mu: The Gaussian Process mean.
    :type mu: int
    :param L: A matrix such that :math:`L L^\top \approx K`, where :math:`K`
        is the covariance matrix.
    :type L: array-like
    :return: initial_value - The argmin :math:`z`.
    :rtype: array-like
    """
    target = mle(nn_distances, d) - mu
    return Ridge(fit_intercept=False).fit(L, target).coef_


def compute_initial_dimensionalities(x, mu_dim, mu_dens, L, nn_distances, d):
    R"""
    Computes an initial guess for the log dimensionality and log density at every cell state
    with Ridge regression.

    :param x: The cell states.
    :type x: array-like
    :param mu: The Gaussian Process mean.
    :type mu: int
    :param L: A matrix such that :math:`L L^\top \approx K`, where :math:`K`
        is the covariance matrix.
    :type L: array-like
    :param nn_distances: The observed nearest neighbor distances.
    :type nn_distances: array-like
    :param d: The local dimensionality of the data.
    :type d: array-like
    :return: initial_value
    :rtype: array-like
    """
    target = log(d) - mu_dim
    initial_dims = Ridge(fit_intercept=False).fit(L, target).coef_
    initial_dens = compute_initial_value(nn_distances, d, mu_dens, L)
    initial_value = stack([initial_dims, initial_dens])
    return initial_value



from .decomposition import DEFAULT_RANK, DEFAULT_METHOD
from .base_model import BaseEstimator, DEFAULT_COV_FUNC
from .inference import (
    compute_transform,
    compute_loss_func,
    compute_log_density_x,
    compute_conditional_mean,
    DEFAULT_N_ITER,
    DEFAULT_INIT_LEARN_RATE,
    DEFAULT_JIT,
    DEFAULT_OPTIMIZER,
)
from .parameters import (
    compute_d,
    compute_d_factal,
    compute_mu,
    compute_initial_value,
    DEFAULT_N_LANDMARKS,
)
from .util import (
    DEFAULT_JITTER,
    Log,
)
from .validation import (
    _validate_string,
    _validate_array,
)


DEFAULT_D_METHOD = "embedding"

logger = Log()


class DensityEstimator(BaseEstimator):
    R"""
    A class for non-parametric density estimation. It performs Bayesian inference with
    a Gaussian process prior and Nearest Neighbor likelihood. All intermediate computations
    are cached as instance variables, which allows viewing intermediate results and
    saving computation time by passing precomputed values as arguments to a new model.

    Parameters
    ----------
    cov_func_curry : function or type
        The generator of the Gaussian process covariance function.
        This must be a curry that takes one length scale argument and returns a
        covariance function of the form k(x, y) :math:`\rightarrow` float.
        Defaults to Matern52.

    n_landmarks : int
        The number of landmark points. If less than 1 or greater than or equal to the
        number of training points, inducing points will not be computed or used.
        Defaults to 5000.

    rank : int or float
        The rank of the approximate covariance matrix. If rank is an int, an :math:`n \times`
        rank matrix :math:`L` is computed such that :math:`L L^\top \approx K`, where `K` is the
        exact :math:`n \times n` covariance matrix. If rank is a float 0.0 :math:`\le` rank
        :math:`\le` 1.0, the rank/size of :math:`L` is selected such that the included eigenvalues
        of the covariance between landmark points account for the specified percentage of the sum
        of eigenvalues. Defaults to 0.99.

    method : str
        Determines how the rank is interpreted: as a fixed number of eigenvectors ('fixed'), a
        percent of eigenvalues ('percent'), or automatically ('auto'). If 'auto', the rank is
        interpreted as a fixed number of eigenvectors if it is an int and as a percent of
        eigenvalues if it is a float. This parameter is provided for clarity in the ambiguous case
        of 1 vs 1.0. Defaults to 'auto'.

    d_method : str
        The method to compute the intrinsic dimensionality of the data. Implemented options are
         - 'embedding': uses the embedding dimension `x.shape[1]`
         - 'fractal': uses the average fractal dimension (experimental)

        Defaults to 'embedding'.

    jitter : float
        A small amount added to the diagonal of the covariance matrix to bind eigenvalues
        numerically away from 0, ensuring numerical stability. Defaults to 1e-6.

    optimizer : str
        The optimizer for the maximum a posteriori or posterior density estimation. Options are
        'L-BFGS-B', stochastic optimizer 'adam', or automatic differentiation variational
        inference 'advi'. Defaults to 'L-BFGS-B'.

    n_iter : int
        The number of optimization iterations. Defaults to 100.

    init_learn_rate : float
        The initial learning rate. Defaults to 1.

    landmarks : array-like or None
        The points used to quantize the data for the approximate covariance. If None,
        landmarks are set as k-means centroids with k=n_landmarks. This is ignored if n_landmarks
        is greater than or equal to the number of training points. Defaults to None.

    nn_distances : array-like or None
        The nearest neighbor distances at each data point. If None, the nearest neighbor
        distances are computed automatically, using a KDTree if the dimensionality of the data
        is less than 20, or a BallTree otherwise. Defaults to None.

    d : int, array-like or None
        The intrinsic dimensionality of the data, i.e., the dimensionality of the embedded
        manifold. If None, `d` is set to the size of axis 1 of the training data points.
        Defaults to None.

    mu : float or None
        The mean :math:`\mu` of the Gaussian process. If None, sets :math:`\mu` to the 1st
        percentile of :math:`\text{mle}(\text{nn_distances}, d) - 10`, where :math:`\text{mle} =
        \log(\text{gamma}(d/2 + 1)) - (d/2) \cdot \log(\pi) - d \cdot \log(\text{nn_distances})`.
        Defaults to None.

    ls : float or None
        The length scale of the Gaussian process covariance function. If None, `ls` is set to
        the geometric mean of the nearest neighbor distances times a constant. If `cov_func`
        is supplied explicitly, `ls` has no effect. Defaults to None.

    cov_func : function or None
        The Gaussian process covariance function of the form k(x, y) :math:`\rightarrow` float.
        If None, the covariance function `cov_func` is automatically generated as `cov_func_curry(ls)`.
        Defaults to None.

    L : array-like or None
        A matrix such that :math:`L L^\top \approx K`, where :math:`K` is the covariance matrix.
        If None, `L` is computed automatically. Defaults to None.

    initial_value : array-like or None
        The initial guess for optimization. If None, the value :math:`z` that minimizes
        :math:`||Lz + \mu - mle|| + ||z||` is found, where :math:`\text{mle} = \log(\text{gamma}(d/2 + 1))
        - (d/2) \cdot \log(\pi) - d \cdot \log(\text{nn_distances})` and :math:`d` is the intrinsic
        dimensionality of the data. Defaults to None.

    jit : bool
        Use jax just-in-time compilation for loss and its gradient during optimization.
        Defaults to False.
    """

    def __init__(
        self,
        cov_func_curry=DEFAULT_COV_FUNC,
        n_landmarks=DEFAULT_N_LANDMARKS,
        rank=DEFAULT_RANK,
        method=DEFAULT_METHOD,
        d_method=DEFAULT_D_METHOD,
        jitter=DEFAULT_JITTER,
        optimizer=DEFAULT_OPTIMIZER,
        n_iter=DEFAULT_N_ITER,
        init_learn_rate=DEFAULT_INIT_LEARN_RATE,
        landmarks=None,
        nn_distances=None,
        d=None,
        mu=None,
        ls=None,
        ls_factor=1,
        cov_func=None,
        L=None,
        initial_value=None,
        jit=DEFAULT_JIT,
    ):
        super().__init__(
            cov_func_curry=cov_func_curry,
            n_landmarks=n_landmarks,
            rank=rank,
            jitter=jitter,
            optimizer=optimizer,
            n_iter=n_iter,
            init_learn_rate=init_learn_rate,
            landmarks=landmarks,
            nn_distances=nn_distances,
            d=d,
            mu=mu,
            ls=ls,
            ls_factor=ls_factor,
            cov_func=cov_func,
            L=L,
            initial_value=initial_value,
            jit=jit,
        )
        self.d_method = _validate_string(
            d_method, "d_method", choices={"fractal", "embedding"}
        )
        self.transform = None
        self.loss_func = None
        self.opt_state = None
        self.losses = None
        self.pre_transformation = None
        self.log_density_x = None
        self.log_density_func = None

    def __repr__(self):
        name = self.__class__.__name__
        string = (
            f"{name}("
            f"cov_func_curry={self.cov_func_curry}, "
            f"n_landmarks={self.n_landmarks}, "
            f"rank={self.rank}, "
            f"method='{self.method}', "
            f"jitter={self.jitter}, "
            f"optimizer='{self.optimizer}', "
            f"n_iter={self.n_iter}, "
            f"init_learn_rate={self.init_learn_rate}, "
            f"landmarks={self.landmarks}, "
        )
        if self.nn_distances is None:
            string += "nn_distances=None, "
        else:
            string += "nn_distances=nn_distances, "
        string += (
            f"d={self.d}, "
            f"mu={self.mu}, "
            f"ls={self.ls}, "
            f"cov_func={self.cov_func}, "
        )
        if self.L is None:
            string += "L=None, "
        else:
            string += "L=L, "
        if self.initial_value is None:
            string += "initial_value=None, "
        else:
            string += "initial_value=initial_value, "
        string += f"jit={self.jit}" ")"
        return string

    def _compute_d(self):
        x = self.x
        if self.d_method == "fractal":
            logger.warning("Using EXPERIMENTAL fractal dimensionality selection.")
            d = compute_d_factal(x)
            logger.info(f"Using d={d}.")
        else:
            d = compute_d(x)
        if d > 50:
            message = f"""The detected dimensionality of the data is over 50,
            which is likely to cause numerical instability issues.
            Consider running a dimensionality reduction algorithm, or
            if this number of dimensions is intended, explicitly pass
            d={self.d} as a parameter."""
            raise ValueError(message)
        return d

    def _compute_mu(self):
        nn_distances = self.nn_distances
        d = self.d
        mu = compute_mu(nn_distances, d)
        return mu

    def _compute_initial_value(self):
        nn_distances = self.nn_distances
        d = self.d
        mu = self.mu
        L = self.L
        initial_value = compute_initial_value(nn_distances, d, mu, L)
        return initial_value

    def _compute_transform(self):
        mu = self.mu
        L = self.L
        transform = compute_transform(mu, L)
        return transform

    def _compute_loss_func(self):
        nn_distances = self.nn_distances
        d = self.d
        transform = self.transform
        k = self.initial_value.shape[0]
        loss_func = compute_loss_func(nn_distances, d, transform, k)
        return loss_func

    def _set_log_density_x(self):
        pre_transformation = self.pre_transformation
        transform = self.transform
        log_density_x = compute_log_density_x(pre_transformation, transform)
        self.log_density_x = log_density_x

    def _set_log_density_func(self):
        x = self.x
        landmarks = self.landmarks
        pre_transformation = self.pre_transformation
        log_density_x = self.log_density_x
        mu = self.mu
        cov_func = self.cov_func
        jitter = self.jitter
        logger.info("Computing predictive function.")
        log_density_func = compute_conditional_mean(
            x,
            landmarks,
            pre_transformation,
            log_density_x,
            mu,
            cov_func,
            jitter=jitter,
        )
        self.log_density_func = log_density_func

    def prepare_inference(self, x):
        R"""
        Set all attributes in preparation for optimization, but do not
        perform Bayesian inference. It is not necessary to call this
        function before calling fit.

        :param x: The training instances to estimate density function.
        :type x: array-like
        :return: loss_func, initial_value - The Bayesian loss function and
            initial guess for optimization.
        :rtype: function, array-like
        """
        if x is None:
            x = self.x
            if self.x is None:
                message = "Required argument x is missing and self.x has not been set."
                raise ValueError(message)
        else:
            if self.x is not None and self.x is not x:
                message = (
                    "self.x has been set already, but is not equal to the argument x."
                )
                raise ValueError(message)

        x = self.set_x(x)
        self._prepare_attribute("nn_distances")
        self._prepare_attribute("d")
        self._prepare_attribute("mu")
        self._prepare_attribute("ls")
        self._prepare_attribute("cov_func")
        self._prepare_attribute("landmarks")
        self._prepare_attribute("L")
        self._prepare_attribute("initial_value")
        self._prepare_attribute("transform")
        self._prepare_attribute("loss_func")
        return self.loss_func, self.initial_value

    def run_inference(self, loss_func=None, initial_value=None, optimizer=None):
        R"""
        Perform Bayesian inference, optimizing the pre_transformation parameters.
        If you would like to run your own inference procedure, use the loss_function
        and initial_value attributes and set pre_transformation to the optimized
        parameters.

        :param loss_func: The Bayesian loss function. If None, uses the stored
            loss_func attribute.
        :type loss_func: function
        :param initial_value: The initial guess for optimization. If None, uses
            the stored initial_value attribute.
        :type initial_value: array-like
        :return: pre_transformation - The optimized parameters.
        :rtype: array-like
        """
        if loss_func is not None:
            self.loss_func = loss_func
        if initial_value is not None:
            self.initial_value = initial_value
        if optimizer is not None:
            self.optimizer = optimizer
        self._run_inference()
        return self.pre_transformation

    def process_inference(self, pre_transformation=None, build_predict=True):
        R"""
        Use the optimized parameters to compute the log density at the
        training points. If build_predict, also build the prediction function.

        :param pre_transformation: The optimized parameters. If None, uses the stored
            pre_transformation attribute.
        :type pre_transformation: array-like
        :param build_predict: Whether or not to build the prediction function.
            Defaults to True.
        :type build_predict: bool
        :return: log_density_x - The log density
        :rtype: array-like
        """
        if pre_transformation is not None:
            self.pre_transformation = _validate_array(
                pre_transformation, "pre_transformation"
            )
        self._set_log_density_x()
        if build_predict:
            self._set_log_density_func()
        return self.log_density_x

    def fit(self, x=None, build_predict=True):
        """
        Trains the model from end to end. This includes preparing the model for inference,
        running the inference, and post-processing the inference results.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features), default=None
            The training instances where `n_samples` is the number of samples and `n_features`
            is the number of features.

        build_predict : bool, default=True
            Whether to build the prediction function after training.

        Returns
        -------
        self : object
            This method returns self for chaining.
        """
        self.prepare_inference(x)
        self.run_inference()
        self.process_inference(build_predict=build_predict)
        return self

    @property
    def predict(self):
        """
        A property that returns an instance of the :class:`mellon.Predictor` class. This predictor can
        be used to predict the log density for new data points by calling the instance like a function.

        The predictor instance also supports serialization features, which allow for saving and loading
        the predictor's state. For more details, refer to the :class:`mellon.Predictor` documentation.

        Returns
        -------
        mellon.Predictor
            A predictor instance that computes the log density at each new data point.

        Example
        -------

        >>> log_density = model.predict(Xnew)

        """
        if self.log_density_func is None:
            self._set_log_density_func()
        return self.log_density_func

    def fit_predict(self, x=None, build_predict=False):
        """
        Trains the model and predicts the log density at the training points.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_features), default=None
            The training instances where `n_samples` is the number of samples and `n_features`
            is the number of features.

        build_predict : bool, default=False
            Whether to build the prediction function after training.

        Raises
        ------
        ValueError
            If the input `x` is not consistent with the training data used before.

        Returns
        -------
        array-like
            The log density at each training point in `x`.
        """
        if self.x is not None and self.x is not x:
            message = "self.x has been set already, but is not equal to the argument x."
            error = ValueError(message)
            logger.error(error)
            raise error
        if self.x is None and x is None:
            message = "Required argument x is missing and self.x has not been set."
            error = ValueError(message)
            logger.error(error)
            raise error
        if x is None:
            x = self.x
        else:
            x = _validate_array(x, "x")

        self.fit(x, build_predict=build_predict)
        return self.log_density_x



import sys
import logging
import functools
import inspect
from inspect import Parameter

from jax.config import config as jaxconfig
from jax.numpy import (
    eye,
    log,
    pi,
    repeat,
    newaxis,
    tensordot,
    sqrt,
    maximum,
    triu_indices,
    sort,
    ones,
    arange,
    concatenate,
    atleast_2d,
    ndarray,
    array,
    isscalar,
    exp,
)
from jax.numpy import sum as arraysum
from jax.numpy.linalg import norm, lstsq, matrix_rank
from jax.scipy.special import gammaln
from jax import vmap, jit
from sklearn.neighbors import BallTree, KDTree

from .validation import _validate_array

DEFAULT_JITTER = 1e-6
DEFAULT_RANK_TOL = 5e-1


def Exp(func):
    """
    Function wrapper, making a function that returns the exponent of the wrapped function.
    """

    def new_func(x):
        return exp(func(x))

    return new_func


def make_serializable(x):
    """
    Convert the input into a serializable format.

    Parameters
    ----------
    x : variable
        The input variable that can be array-like, slice, scalar or dict.

    Returns
    -------
    serializable_x : variable
        The input variable converted into a serializable format.
    """
    if isinstance(x, ndarray):
        return {"type": "jax.numpy", "data": x.tolist()}
    elif isinstance(x, slice):
        return {"type": "slice", "data": (x.start, x.stop, x.step)}
    elif isinstance(x, dict):
        return {"type": "dict", "data": {k: make_serializable(v) for k, v in x.items()}}
    else:
        return x


def deserialize(serializable_x):
    """
    Convert the serializable input back into the original format.

    Parameters
    ----------
    serializable_x : variable
        The input variable that is in a serializable format.

    Returns
    -------
    x : variable
        The input variable converted back into its original format.
    """
    if isinstance(serializable_x, dict):
        data_type = serializable_x["type"]
        if data_type == "jax.numpy":
            return array(serializable_x["data"])
        elif data_type == "slice":
            return slice(*serializable_x["data"])
        elif data_type == "dict":
            return {k: deserialize(v) for k, v in serializable_x["data"].items()}
    else:
        return serializable_x


def ensure_2d(X):
    """
    Ensures that the input JAX array, X, is at least 2-dimensional.

    :param X: The input JAX array to be made 2-dimensional.
    :type X: jnp.array
    :return: The input array transformed to a 2-dimensional array.
    :rtype: jnp.array

    If X is 1-dimensional, it is reshaped to a 2-dimensional array,
    where each element of X becomes a row in the 2-dimensional array.
    """
    return atleast_2d(X.T).T


def select_active_dims(x, active_dims):
    """
    Select the active dimensions from the input.

    Parameters
    ----------
    x : array-like
        Input array.

    selected_dimensions : array-like, slice or scalar
        The indices of the active dimensions. It could be a scalar, a list, a numpy array, or a slice object.

    Returns
    -------
    x : array-like
        Array with selected dimensions.
    """
    if active_dims is not None:
        if isscalar(active_dims):
            active_dims = [active_dims]
        x = x[:, active_dims]
    return x


def make_multi_time_argument(func):
    """
    Decorator to modify a method to optionally take a multi-time argument.

    This decorator modifies the method it wraps to take an optional `multi_time` keyword argument.
    If `multi_time` is provided, the decorated method will be called once for each value in `multi_time`
    with that value passed as the `time` argument to the method.

    The original method's signature and docstring are preserved.

    Parameters
    ----------
    func : callable
        The method to be modified. This method must take a `time` keyword argument.

    Returns
    -------
    callable
        The modified method.

    Examples
    --------
    class MyClass:
        @make_multi_time_argument
        def method(self, x, time=None):
            return x + time

    my_object = MyClass()
    print(my_object.method(1, multi_time=np.array([1, 2, 3])))
    # Output: array([2, 3, 4])
    """
    sig = inspect.signature(func)
    new_params = list(sig.parameters.values()) + [
        Parameter("multi_time", Parameter.POSITIONAL_OR_KEYWORD, default=None)
    ]
    new_sig = sig.replace(parameters=new_params)

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        multi_time = kwargs.pop("multi_time", None)
        do_jit = kwargs.get("jit", False)
        if multi_time is not None:
            if kwargs.get("time", None) is not None:
                raise ValueError(
                    "Cannot specify both 'time' and 'multi_time' arguments"
                )
            multi_time = _validate_array(multi_time, "multi_time")

            def at_time(t):
                return func(self, *args, **kwargs, time=t)

            if do_jit:
                at_time = jit(at_time)
            vfun = vmap(at_time, in_axes=0, out_axes=1)
            return vfun(multi_time)
        return func(self, *args, **kwargs)

    wrapper.__signature__ = new_sig
    return wrapper


def stabilize(A, jitter=DEFAULT_JITTER):
    R"""
    Add a small jitter to the diagonal for numerical stability.

    :param A: A square matrix.
    :param jitter: The amount to add to the diagonal. Defaults to 1e-6.
    :type jitter: float
    :return: :math:`A'` - The matrix :math:`A` with a small jitter added to the diagonal.
    :rtype: array-like
    """
    n = A.shape[0]
    return A + eye(n) * jitter


def mle(nn_distances, d):
    R"""
    Nearest Neighbor distribution maximum likelihood estimate for log density
    given observed nearest neighbor distances :math:`nn\text{_}distances` in
    dimensions :math:`d`: :math:`mle = \log(\text{gamma}(d/2 + 1)) - (d/2) \cdot \log(\pi) -
    d \cdot \log(nn\text{_}distances)`

    :param nn_distances: The observed nearest neighbor distances.
    :type nn_distances: array-like
    :param d: The local dimensionality of the data.
    :type d: int
    :return: :math:`mle` - The maximum likelihood estimate at each point.
    :rtype: array-like
    """
    return gammaln(d / 2 + 1) - (d / 2) * log(pi) - d * log(nn_distances)


def distance(x, y):
    """
    Computes the distances between each point in x and y.

    :param x: A set of points.
    :type x: array-like
    :param y: A set of points.
    :type y: array-like
    :return: distances - The distance between each point in x and y.
    :rtype: array-like
    """
    n = x.shape[0]
    m = y.shape[0]
    xx = repeat(arraysum(x * x, axis=1)[:, newaxis], m, axis=1)
    yy = repeat(arraysum(y * y, axis=1)[newaxis, :], n, axis=0)
    xy = tensordot(x, y, (1, 1))
    sq = xx - 2 * xy + yy + 1e-12
    return sqrt(maximum(sq, 0))


def test_rank(input, tol=DEFAULT_RANK_TOL, threshold=None):
    """
    Inspects the approximate rank of the transformation matrix L. The input can be the matrix itself or an object
    containing L as an attribute. A high rank indicates a potentially insufficient latent representation,
    suggesting a need for a more complex transformation matrix. Also allows logging based on a rank fraction threshold.

    :param input: The matrix L or an object containing it as an attribute.
    :type input: array-like or mellon estimator object
    :param tol: The rank calculation tolerance, defaults to {DEFAULT_RANK_TOL}.
    :type tol: float, optional
    :param threshold: If provided, logs a message based on the rank fraction.
    :type threshold: float, optional
    :return: The approximate rank of the matrix.
    :rtype: int
    :raises ValueError: If the input matrix is not 2D.
    """
    if hasattr(input, "shape"):
        L = input
    elif hasattr(input, "L"):
        L = input.L
        if L is None:
            raise AttributeError(
                "Matrix L is not found in the estimator object. Consider running `.prepare_inference()`."
            )
    else:
        raise TypeError(
            "Input must be either a matrix or a mellon enstimator with a transformation L."
        )

    if len(L.shape) != 2:
        raise ValueError("Matrix L must be 2D.")

    approx_rank = matrix_rank(L, tol=tol)
    max_rank = min(L.shape)
    rank_fraction = approx_rank / max_rank

    if threshold is not None:
        logger = Log()
        if rank_fraction > threshold:
            logger.warning(
                f"High approx. rank fraction ({rank_fraction:.1%}). "
                "Consider increasing 'n_landmarks'."
            )
        else:
            logger.info(
                f"Rank fraction ({rank_fraction:.1%}, lower is better) is "
                "within acceptable range. "
                "Current settings should provide satisfactory model performance."
            )
    else:
        print(
            f"The approx. rank fraction is {rank_fraction:.1%} "
            f"({approx_rank:,} of {max_rank:,}). Lower is better."
        )

    return approx_rank.item()


def local_dimensionality(x, k=30, x_query=None, neighbor_idx=None):
    """
    Compute an estimate of the local fractal dimension of a data set using nearest neighbors.

    :param x: The input samples.
    :type x: array-like of shape (n_samples, n_features)
    :param k: The number of neighbors to consider, defaults to 30.
    :type k: int, optional
    :param x_query: The points at which to compute the local fractal dimension.
        If None, use x itself, defaults to None.
    :type x_query: array-like of shape (n_queries, n_features), optional
    :param neighbor_idx: The indices of the neighbors for each query point.
        If None, these are computed using a nearest neighbor search, defaults to None.
    :type neighbor_idx: array-like of shape (n_queries, k), optional
    :return: The estimated local fractal dimension at each query point.
    :rtype: array-like of shape (n_queries,)

    This function computes the local fractal dimension of a dataset at query points.
    It uses nearest neighbors and fits a line in log-log space to estimate the fractal dimension.
    """
    if neighbor_idx is None:
        if x_query is None:
            x_query = x
        if x.shape[1] >= 20:
            tree = BallTree(x, metric="euclidean")
        else:
            tree = KDTree(x, metric="euclidean")
        neighbors = x[tree.query(x_query, k=k)[1]]
    else:
        neighbors = x[neighbor_idx]
    i, j = triu_indices(k, k=1)
    neighbor_distances = norm(neighbors[..., i, :] - neighbors[..., j, :], axis=-1)
    neighborhood_distances = sort(neighbor_distances, axis=-1)

    kc2 = k * (k - 1) // 2
    A = concatenate(
        [log(neighborhood_distances)[..., None], ones((x_query.shape[0], kc2, 1))],
        axis=-1,
    )
    y = log(arange(1, kc2 + 1))[:, None]

    vreg = vmap(lstsq, in_axes=(0, None))
    w, _, _, _ = vreg(A, y)
    return w[:, 0, 0]


class Log(object):
    """Access the Mellon logging/verbosity. Log() returns the singelon logger and
    Log.off() and Log.on() disable or enable logging respectively."""

    def __new__(cls):
        """Return the singelton Logger."""
        if not hasattr(cls, "logger"):
            logger = logging.getLogger(__name__)
            logger.setLevel(logging.INFO)
            cls.handler = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter("[%(asctime)s] [%(levelname)-8s] %(message)s")
            cls.handler.setFormatter(formatter)
            logger.addHandler(cls.handler)
            logger.propagate = False
            cls.logger = logger
        return cls.logger

    @classmethod
    def off(cls):
        """Turn off all logging."""
        cls.__new__(cls)
        cls.logger.setLevel(logging.WARNING + 1)

    @classmethod
    def on(cls):
        """Turn on logging."""
        cls.__new__(cls)
        cls.logger.setLevel(logging.INFO)


def set_jax_config(enable_x64=True, platform_name="cpu"):
    """
    Sets up the JAX configuration with the specified settings.

    Parameters
    ----------
    enable_x64 : bool, optional
        Whether to enable 64-bit (double precision) computations in JAX.
        Defaults to True.
    platform_name : str, optional
        The platform name to use in JAX ('cpu', 'gpu', or 'tpu').
        Defaults to 'cpu'.
    """
    jaxconfig.update("jax_enable_x64", enable_x64)
    jaxconfig.update("jax_platform_name", platform_name)
